#!/usr/bin/env python
import hashlib
import shutil
import tempfile
from contextlib import (
    contextmanager,
    nullcontext,
)
from functools import partial
from pathlib import Path
from subprocess import Popen
from urllib.parse import urlparse

import toolz
import yaml
from attr import (
    field,
    frozen,
)
from attr.validators import (
    deep_iterable,
    instance_of,
    optional,
)
from git import (
    Blob,
    Remote,
    Repo,
)

from xorq.catalog.constants import (
    CATALOG_YAML_NAME,
    METADATA_APPEND,
    PREFERRED_SUFFIX,
    VALID_SUFFIXES,
    CatalogInfix,
)
from xorq.catalog.expr_utils import (
    build_expr_context,
    build_expr_context_tgz,
    load_expr_from_tgz,
)
from xorq.catalog.git_utils import (
    add_as_submodule,
    commit_context,
)
from xorq.catalog.tar_utils import (
    make_tgz_context,
    test_tgz,
)


abspath = toolz.compose(Path.absolute, Path)
popen_shell = partial(Popen, shell=True)


def with_pure_suffix(path, suffix=""):
    return path.with_name(path.name.removesuffix("".join(path.suffixes))).with_suffix(
        suffix
    )


@frozen
class Catalog:
    repo = field(validator=instance_of(Repo))

    by_name_base_path = Path("~/.local/share/xorq/git-catalogs").expanduser()
    submodule_rel_path = Path(".xorq/git-catalogs")

    def __attrs_post_init__(self):
        self._ensure_catalog_yaml()
        self.assert_consistency()

    def _ensure_catalog_yaml(self):
        assert not self.repo.bare
        if not any(
            self.catalog_yaml.yaml_relpath.name == blob.name
            for blob in self.repo.head.commit.tree.list_traverse()
        ):
            with self.commit_context(f"add {CATALOG_YAML_NAME}") as index:
                index.add(self.catalog_yaml.yaml_path)

    @property
    def repo_path(self):
        return Path(self.repo.working_dir)

    @property
    def catalog_yaml(self):
        return CatalogYAML(self.repo_path)

    def _add_tgz(self, path, sync=True, aliases=()):
        # should we enable not syncing?
        with self.maybe_synchronizing(sync):
            catalog_addition = CatalogAddition(BuildTgz(path), self, aliases=aliases)
            catalog_entry = catalog_addition.add()
            self.assert_consistency()
            return catalog_entry

    def _add_build_dir(self, build_dir, sync=True, aliases=()):
        with make_tgz_context(build_dir) as tgz_path:
            return self._add_tgz(tgz_path, sync=sync, aliases=aliases)

    def _add_expr(self, expr, sync=True, aliases=()):
        with build_expr_context(expr) as path:
            return self._add_build_dir(path, sync=sync, aliases=aliases)

    def add(self, obj, sync=True, aliases=()):
        from xorq.api import Expr

        match obj:
            case Path() if obj.is_dir():
                f = self._add_build_dir
            case Path() if obj.is_file():
                f = self._add_tgz
            case Expr():
                f = self._add_expr
            case _:
                raise ValueError(f"don't know how to handle type={type(obj)}")
        return f(obj, sync=sync, aliases=aliases)

    def remove(self, name, sync=True):
        with self.maybe_synchronizing(sync):
            catalog_removal = CatalogRemoval.from_name_catalog(name, self)
            catalog_entry = catalog_removal.remove()
            self.assert_consistency()
            return catalog_entry

    def list(self):
        return self.catalog_yaml.contents[CatalogInfix.ENTRY]

    def fetch(self):
        return tuple(map(Remote.fetch, self.repo.remotes))

    def push(self):
        self.assert_consistency()
        return tuple(map(Remote.push, self.repo.remotes))

    def pull(self):
        return tuple(map(Remote.pull, self.repo.remotes))

    @contextmanager
    def synchronizing(self):
        self.pull()
        yield
        self.push()

    @contextmanager
    def maybe_synchronizing(self, sync):
        ctx = self.synchronizing if sync else nullcontext
        with ctx():
            yield

    def sync(self):
        with self.synchronizing():
            pass

    def contains(self, name):
        catalog_entry = CatalogEntry(name, self, require_exists=False)
        return catalog_entry.exists()

    def get_catalog_entry(self, name):
        assert name in self.list(), f"Entry '{name}' not found in catalog"
        catalog_entry = CatalogEntry(name, self)
        return catalog_entry

    def get_tgz(self, name, dir_path=None):
        catalog_entry = self.get_catalog_entry(name)
        return catalog_entry.get(dir_path)

    @property
    def catalog_entries(self):
        return tuple(CatalogEntry(name, self) for name in self.list())

    @contextmanager
    def commit_context(self, message):
        with commit_context(self.repo, message) as index:
            yield index

    def add_alias(self, name, alias, sync=True):
        with self.maybe_synchronizing(sync):
            catalog_entry = CatalogEntry(name, self)
            catalog_alias = CatalogAlias(alias=alias, catalog_entry=catalog_entry)
            catalog_alias.add()
            return catalog_alias

    def list_aliases(self):
        return self.catalog_yaml.contents[CatalogInfix.ALIAS]

    @property
    def catalog_aliases(self):
        return tuple(
            CatalogAlias(
                alias=alias,
                catalog_entry=CatalogEntry(
                    self.repo_path.joinpath(
                        CatalogInfix.ALIAS, alias + PREFERRED_SUFFIX
                    )
                    .resolve()
                    .with_suffix("")
                    .name,
                    self,
                ),
            )
            for alias in self.list_aliases()
        )

    def assert_consistency(self):
        # catalog_yaml is in repo
        catalog_yaml_relpath_string = str(self.catalog_yaml.yaml_relpath)
        path_strings = tuple(
            blob.path
            for blob in self.repo.head.commit.tree.list_traverse()
            if isinstance(blob, Blob)
        )
        assert catalog_yaml_relpath_string in path_strings

        # yaml aliases match filesystem aliases
        assert sorted(self.catalog_yaml.contents[CatalogInfix.ALIAS]) == sorted(
            ca.alias for ca in self.catalog_aliases
        )

        # everything else in repo is either catalog_path or metadata_path from an entry the catalog_yaml knows about, or an alias symlink
        actual = sorted(el for el in path_strings if el != catalog_yaml_relpath_string)
        expected = sorted(
            (
                *(
                    str(path.relative_to(self.repo_path))
                    for catalog_entry in self.catalog_entries
                    for path in (
                        catalog_entry.metadata_path,
                        catalog_entry.catalog_path,
                    )
                ),
                *(
                    str(catalog_alias.alias_path.relative_to(self.repo_path))
                    for catalog_alias in self.catalog_aliases
                ),
            )
        )
        assert actual == expected

    def add_as_submodule(self, root_repo):
        message = f"add submodule: {self.repo_path.name}"
        with commit_context(root_repo, message):
            add_as_submodule(root_repo, self.repo)

    @classmethod
    def clone_from(cls, url, repo_path=None):
        if repo_path is None:
            name = Path(urlparse(url).path).stem
            repo_path = cls.name_to_repo_path(name)
        repo = Repo.clone_from(url, repo_path)
        return cls(repo=repo)

    @classmethod
    def from_repo_path(cls, repo_path, init=None):
        init = not Path(repo_path).exists() if init is None else init
        repo = cls.init_repo_path(repo_path) if init else Repo(repo_path)
        return cls(repo=repo)

    @classmethod
    def from_name(cls, name, init=None):
        repo_path = cls.name_to_repo_path(name)
        return cls.from_repo_path(repo_path, init=init)

    @classmethod
    def from_default(cls, init=None):
        return cls.from_name(name="default", init=init)

    @classmethod
    def clone_from_as_submodule(cls, root_repo, url):
        name = Path(urlparse(url).path).stem
        repo_path = Path(root_repo.working_dir).joinpath(cls.submodule_rel_path, name)
        self = cls.clone_from(url, repo_path)
        self.add_as_submodule(root_repo)
        return self

    @classmethod
    def from_name_as_submodule(cls, root_repo, name, init=None):
        repo_path = Path(root_repo.working_dir).joinpath(cls.submodule_rel_path, name)
        self = cls.from_repo_path(repo_path, init=init)
        self.add_as_submodule(root_repo)
        return self

    @classmethod
    def from_kwargs(cls, name=None, path=None, url=None, root_repo=None, init=None):
        if isinstance(root_repo, (str, Path)):
            root_repo = Repo(root_repo)
        if root_repo:
            match (name, url, path):
                case (None, str(), None):
                    return cls.clone_from_as_submodule(root_repo=root_repo, url=url)
                case (str(), None, None):
                    return cls.from_name_as_submodule(root_repo=root_repo, name=name)
                case _:
                    raise ValueError(
                        f"With `root_repo`, provide exactly one of `name` or `url` "
                        f"(not `path`). Got: name={name!r}, url={url!r}, path={path!r}"
                    )
        elif url:
            if name:
                raise ValueError(
                    f"`url` and `name` are mutually exclusive. "
                    f"Got: name={name!r}, url={url!r}"
                )
            else:
                return cls.clone_from(url=url, repo_path=path)
        else:
            match (name, path):
                case (None, None):
                    return cls.from_default(init=init)
                case (str(), None):
                    return cls.from_name(name=name, init=init)
                case (None, str() | Path()):
                    catalog = Catalog.from_repo_path(Path(path), init=init)
                case _:
                    raise ValueError("`name` and `path` are mutually exclusive.")
            return catalog

    @classmethod
    def name_to_repo_path(cls, name):
        repo_path = cls.by_name_base_path.joinpath(name)
        return repo_path

    @staticmethod
    def init_repo_path(repo_path, bare=False):
        assert not (repo_path := Path(repo_path)).exists(), (
            f"Catalog repo already exists at {repo_path}"
        )
        repo = Repo.init(repo_path, mkdir=True, bare=bare)
        repo.index.commit("initial commit")
        return repo


@frozen
class BuildTgz:
    path = field(validator=instance_of(Path), converter=Path)

    def __attrs_post_init__(self):
        assert "".join(self.path.suffixes) in VALID_SUFFIXES, (
            f"Invalid archive suffix '{self.path.suffixes}', expected one of {VALID_SUFFIXES}"
        )
        assert self.path.exists(), f"Build archive not found at {self.path}"
        test_tgz(self.path)

    @property
    def name(self):
        return with_pure_suffix(self.path, "").name

    @property
    def md5sum(self):
        from xorq.common.utils.dask_normalize.dask_normalize_utils import file_digest

        return file_digest(self.path, hashlib.md5)


@frozen
class CatalogAddition:
    build_tgz = field(validator=instance_of(BuildTgz))
    catalog = field(validator=instance_of(Catalog))
    aliases = field(validator=deep_iterable(instance_of(str)), default=())
    _maybe_tmpfile = field(
        validator=optional(instance_of(tempfile._TemporaryFileWrapper)),
        default=None,
    )

    @property
    def name(self):
        return self.build_tgz.name

    @property
    def metadata(self):
        return {"md5sum": self.build_tgz.md5sum}

    @property
    def catalog_entry(self):
        return CatalogEntry(self.name, self.catalog, require_exists=False)

    def ensure_dirs(self):
        for p in (self.catalog_entry.metadata_path, self.catalog_entry.catalog_path):
            p.parent.mkdir(exist_ok=True, parents=True)

    @property
    def catalog_aliases(self):
        return tuple(CatalogAlias(alias, self.catalog_entry) for alias in self.aliases)

    @property
    def message(self):
        alias_message = f" (aliases {', '.join(self.aliases)})" if self.aliases else ""
        message = f"add: {self.name}{alias_message}"
        return message

    def _add(self):
        assert not self.catalog.contains(self.name), (
            f"Entry '{self.name}' already exists in catalog"
        )
        self.ensure_dirs()
        catalog_entry = self.catalog_entry
        catalog_entry.metadata_path.write_text(yaml.safe_dump(self.metadata))
        shutil.copy(self.build_tgz.path, catalog_entry.catalog_path)
        index = self.catalog.repo.index
        #
        self.catalog.catalog_yaml.add(self.name)
        index.add(
            (
                catalog_entry.catalog_path,
                catalog_entry.metadata_path,
                self.catalog.catalog_yaml.yaml_path,
            )
        )
        for catalog_alias in self.catalog_aliases:
            catalog_alias._add()
        return CatalogEntry(self.name, self.catalog, require_exists=True)

    def add(self):
        with self.catalog.commit_context(self.message):
            return self._add()

    @classmethod
    def from_expr(cls, expr, catalog):
        ntfh = tempfile.NamedTemporaryFile(suffix=PREFERRED_SUFFIX)
        with build_expr_context_tgz(expr) as tgz_path:
            shutil.copy(tgz_path, ntfh.name)
        return cls(BuildTgz(ntfh.name), catalog, maybe_tmpfile=ntfh)


@frozen
class CatalogEntry:
    name = field(validator=instance_of(str))
    catalog = field(validator=instance_of(Catalog))
    require_exists = field(validator=instance_of(bool), default=True)

    def __attrs_post_init__(self):
        self.assert_consistency()
        if self.require_exists:
            assert self.exists(), f"Catalog entry '{self.name}' does not exist"

    @property
    def repo_path(self):
        return self.catalog.repo_path

    @property
    def metadata_path(self):
        metadata_path = self.repo_path.joinpath(
            CatalogInfix.METADATA, self.name + PREFERRED_SUFFIX + METADATA_APPEND
        )
        return metadata_path

    @property
    def catalog_path(self):
        catalog_path = self.repo_path.joinpath(
            CatalogInfix.ENTRY, self.name
        ).with_suffix(PREFERRED_SUFFIX)
        return catalog_path

    @property
    def expr(self):
        return load_expr_from_tgz(self.catalog_path)

    @property
    def aliases(self):
        return tuple(
            catalog_alias
            for catalog_alias in self.catalog.catalog_aliases
            if catalog_alias.catalog_entry.name == self.name
        )

    @property
    def _exists_components(self):
        return {
            "metadata_path": self.metadata_path.exists(),
            "catalog_path": self.catalog_path.exists(),
            "catalog_yaml_contents": self.catalog.catalog_yaml.contains(self.name),
        }

    def get(self, dir_path=None):
        target = (
            Path(dir_path or Path.cwd())
            .joinpath(self.name)
            .with_suffix(PREFERRED_SUFFIX)
        )
        shutil.copy(self.catalog_path, target)
        return target

    def assert_consistency(self):
        values = self._exists_components.values()
        assert all(values) or not any(values)

    def exists(self):
        return all(self._exists_components.values())


@frozen
class CatalogAlias:
    alias = field(validator=instance_of(str))
    catalog_entry = field(validator=instance_of(CatalogEntry))

    @property
    def alias_path(self):
        return self.catalog_entry.repo_path.joinpath(
            CatalogInfix.ALIAS, self.alias
        ).with_suffix(PREFERRED_SUFFIX)

    @property
    def target(self):
        return Path("..").joinpath(
            CatalogInfix.ENTRY, self.catalog_entry.name + PREFERRED_SUFFIX
        )

    def ensure_dirs(self):
        self.alias_path.parent.mkdir(exist_ok=True, parents=True)

    def _add(self):
        alias_path = self.alias_path
        if alias_path.exists():
            if not alias_path.is_symlink():
                raise ValueError(f"non symlink already exists at {alias_path}")
            alias_path.unlink()
        else:
            self.ensure_dirs()
        alias_path.symlink_to(self.target)
        catalog_yaml = self.catalog_entry.catalog.catalog_yaml
        #
        catalog_yaml.add_alias(self.alias)
        self.catalog_entry.catalog.repo.index.add([alias_path, catalog_yaml.yaml_path])
        return self

    def add(self):
        message = f"add alias: {self.alias} -> {self.catalog_entry.name}"
        with self.catalog_entry.catalog.commit_context(message):
            self._add()

    def list_revisions(self):
        repo = self.catalog_entry.catalog.repo
        catalog = self.catalog_entry.catalog
        alias_relpath = str(self.alias_path.relative_to(self.catalog_entry.repo_path))
        blobs_commits = tuple(
            (
                # commit.tree.join raises KeyError for non-existent paths
                toolz.excepts(KeyError, commit.tree.join)(
                    Path(CatalogInfix.ALIAS, self.alias + PREFERRED_SUFFIX)
                ),
                commit,
            )
            for commit in repo.iter_commits(paths=alias_relpath)
        )
        result = tuple(
            (
                CatalogEntry(
                    with_pure_suffix(Path(blob.data_stream.read().decode()), "").name,
                    catalog,
                    require_exists=False,
                ),
                commit,
            )
            for blob, commit in blobs_commits
            if blob
        )
        return result

    def _remove(self):
        alias_path = self.alias_path
        assert alias_path.is_symlink(), f"no alias symlink at {alias_path}"
        catalog_yaml = self.catalog_entry.catalog.catalog_yaml
        index = self.catalog_entry.catalog.repo.index
        #
        catalog_yaml.remove_alias(self.alias)
        index.add([catalog_yaml.yaml_path])
        index.remove([alias_path])
        alias_path.unlink()
        return self

    def remove(self):
        message = f"rm alias: {self.alias}"
        with self.catalog_entry.catalog.commit_context(message):
            self._remove()
        return self

    @classmethod
    def from_name(cls, name, catalog):
        alias_path = catalog.repo_path.joinpath(
            CatalogInfix.ALIAS,
            name,
        ).with_suffix(PREFERRED_SUFFIX)
        if alias_path.exists():
            catalog_alias = CatalogAlias(
                name,
                CatalogEntry(alias_path.resolve().with_suffix("").name, catalog),
            )
            return catalog_alias
        else:
            raise ValueError(f"no such alias {name}")


@frozen
class CatalogRemoval:
    catalog_entry = field(validator=instance_of(CatalogEntry))

    @property
    def message(self):
        name = with_pure_suffix(self.catalog_entry.catalog_path, "").name
        catalog_aliases = self.catalog_entry.aliases
        aliases_message = (
            f" (aliases {', '.join(catalog_alias.alias for catalog_alias in catalog_aliases)})"
            if catalog_aliases
            else ""
        )
        message = f"rm: {name}{aliases_message}"
        return message

    def _remove(self):
        catalog_entry = self.catalog_entry
        catalog = catalog_entry.catalog
        assert catalog_entry.exists(), (
            f"Cannot remove entry '{catalog_entry.name}': not found in catalog"
        )
        index = catalog.repo.index
        #
        for catalog_alias in self.catalog_entry.aliases:
            catalog_alias._remove()
        catalog.catalog_yaml.remove(catalog_entry.name)
        index.add((catalog.catalog_yaml.yaml_path,))
        paths = (
            catalog_entry.metadata_path,
            catalog_entry.catalog_path,
        )
        index.remove(paths)
        for path in paths:
            path.unlink()
        return catalog_entry

    def remove(self):
        with self.catalog_entry.catalog.commit_context(self.message):
            return self._remove()

    @classmethod
    def from_name_catalog(cls, name, catalog):
        return cls(CatalogEntry(name=name, catalog=catalog))


@frozen
class CatalogYAML:
    repo_path = field(validator=instance_of(Path), converter=abspath)

    def __attrs_post_init__(self):
        if not self.yaml_path.exists():
            self.yaml_path.write_text(
                yaml.safe_dump(
                    {str(CatalogInfix.ENTRY): [], str(CatalogInfix.ALIAS): []}
                )
            )

    @property
    def yaml_path(self):
        return self.repo_path.joinpath(CATALOG_YAML_NAME)

    @property
    def yaml_relpath(self):
        return self.yaml_path.relative_to(self.repo_path)

    @property
    def contents(self):
        raw = yaml.safe_load(self.yaml_path.read_text())
        if isinstance(raw, list):
            # legacy format: plain list of entry names, no aliases section
            return {str(CatalogInfix.ENTRY): raw, str(CatalogInfix.ALIAS): []}
        return raw

    def set_contents(self, contents):
        self.yaml_path.write_text(yaml.safe_dump(contents))
        return self.yaml_path

    def contains(self, entry):
        return entry in self.contents[CatalogInfix.ENTRY]

    def add(self, entry):
        assert not self.contains(entry)
        contents = self.contents
        contents[CatalogInfix.ENTRY] = contents[CatalogInfix.ENTRY] + [entry]
        return self.set_contents(contents)

    def remove(self, entry):
        contents = self.contents
        contents[CatalogInfix.ENTRY] = [
            el for el in contents[CatalogInfix.ENTRY] if el != entry
        ]
        return self.set_contents(contents)

    def contains_alias(self, alias):
        return alias in self.contents[CatalogInfix.ALIAS]

    def add_alias(self, alias):
        if not self.contains_alias(alias):
            contents = self.contents
            contents[CatalogInfix.ALIAS] = contents[CatalogInfix.ALIAS] + [alias]
            self.set_contents(contents)
        return self.yaml_path

    def remove_alias(self, alias):
        contents = self.contents
        contents[CatalogInfix.ALIAS] = [
            el for el in contents[CatalogInfix.ALIAS] if el != alias
        ]
        return self.set_contents(contents)



from contextlib import contextmanager
from functools import partial
from pathlib import Path
from types import SimpleNamespace

import click


def click_handler(e):
    raise click.ClickException(str(e)) from e


def _init_hint(ctx):
    """Return the `xorq catalog ... init` command the user should run."""
    group_ctx = ctx.parent or ctx
    match (group_ctx.params.get("name"), group_ctx.params.get("path")):
        case (str() as name, _):
            return f"xorq catalog --name {name} init"
        case (_, str() as path):
            return f"xorq catalog --path {path} init"
        case _:
            return "xorq catalog init"


@contextmanager
def click_context(*typs):
    try:
        yield
    except click.ClickException:
        raise
    except typs as e:
        click_handler(e)


@contextmanager
def click_context_catalog(ctx):
    from git import NoSuchPathError

    try:
        yield
    except click.ClickException:
        raise
    except NoSuchPathError as e:
        hint = _init_hint(ctx)
        raise click.ClickException(
            f"Catalog not found: {e}\nRun `{hint}` to create it."
        ) from e
    except Exception as e:
        click_handler(e)


click_context_default = partial(click_context, AssertionError, Exception)


def _make_catalog_for_completion(ctx):
    from xorq.catalog.catalog import Catalog

    catalog_ctx = ctx.parent
    return Catalog.from_kwargs(
        name=catalog_ctx.params.get("name"),
        path=catalog_ctx.params.get("path"),
        url=catalog_ctx.params.get("url"),
        root_repo=catalog_ctx.params.get("root_repo"),
        init=False,
    )


def _complete_entry_names(ctx, param, incomplete):
    from click.shell_completion import CompletionItem

    try:
        catalog = _make_catalog_for_completion(ctx)
        return [CompletionItem(n) for n in catalog.list() if n.startswith(incomplete)]
    except Exception:
        return []


def _complete_alias_names(ctx, param, incomplete):
    from click.shell_completion import CompletionItem

    try:
        catalog = _make_catalog_for_completion(ctx)
        return [
            CompletionItem(a)
            for a in catalog.list_aliases()
            if a.startswith(incomplete)
        ]
    except Exception:
        return []


@click.group(invoke_without_command=True)
@click.option(
    "-n", "--name", default=None, help="Catalog name (mutually exclusive with --path)."
)
@click.option(
    "-p",
    "--path",
    default=None,
    type=click.Path(file_okay=False),
    help="Catalog repo path (mutually exclusive with --name).",
)
@click.option(
    "-u",
    "--url",
    default=None,
    help="Remote repo url to clone",
)
@click.option(
    "-r",
    "--root-repo",
    default=None,
    type=click.Path(file_okay=False, exists=True),
    help="Repo root to add this catalog to as a submodule",
)
@click.option(
    "--init/--no-init",
    default=None,
    help="Initialize the repo (default: auto).",
)
@click.pass_context
def cli(ctx, name, path, url, root_repo, init):
    """Manage xorq build-artifact catalogs."""
    from xorq.catalog.catalog import Catalog

    ctx.obj = SimpleNamespace(
        make_catalog=partial(
            Catalog.from_kwargs,
            name=name,
            path=path,
            url=url,
            root_repo=root_repo,
            init=init,
        )
    )


@cli.command()
@click.option("--refresh", default=10, type=float, help="Refresh interval in seconds.")
@click.pass_context
def tui(ctx, refresh):
    """Launch terminal UI."""
    with click_context_catalog(ctx):
        ctx.obj.make_catalog(init=False)  # validate catalog exists
    from xorq.catalog.tui import CatalogTUI

    app = CatalogTUI(
        partial(ctx.obj.make_catalog, init=False), refresh_interval=refresh
    )
    app.run()


@cli.command()
@click.pass_context
def init(ctx):
    """Initialize a new catalog."""
    with click_context_catalog(ctx):
        try:
            catalog = ctx.obj.make_catalog(init=True)
        except AssertionError as err:
            # init_repo_path asserts the path does not already exist
            probe = ctx.obj.make_catalog(init=False)
            raise click.ClickException(
                f"Catalog already exists at {probe.repo_path}"
            ) from err
    click.echo(f"Initialized catalog at {catalog.repo_path}")


@cli.command()
@click.argument("paths", nargs=-1, required=True, type=click.Path(exists=True))
@click.option("--sync/--no-sync", default=True)
@click.option(
    "-a",
    "--alias",
    "aliases",
    multiple=True,
    help="Alias(es) to create for the added entry (repeatable).",
)
@click.pass_context
def add(ctx, paths, sync, aliases):
    """Add entries from tgz files or build directories."""
    with click_context_catalog(ctx):
        catalog = ctx.obj.make_catalog(init=False)
        with catalog.maybe_synchronizing(sync):
            for path in map(Path, paths):
                entry = catalog.add(path, sync=False, aliases=aliases)
                click.echo(f"Added {entry.name}")


@cli.command()
@click.argument("names", nargs=-1, required=True, shell_complete=_complete_entry_names)
@click.option("--sync/--no-sync", default=True)
@click.pass_context
def remove(ctx, names, sync):
    """Remove entries by name."""
    with click_context_catalog(ctx):
        catalog = ctx.obj.make_catalog(init=False)
        with catalog.maybe_synchronizing(sync):
            for name in names:
                entry = catalog.remove(name, sync=False)
                click.echo(f"Removed {entry.name}")


@cli.command("add-alias")
@click.argument("name", shell_complete=_complete_entry_names)
@click.argument("alias")
@click.option("--sync/--no-sync", default=True)
@click.pass_context
def add_alias(ctx, name, alias, sync):
    """Add an alias for an entry."""
    with click_context_catalog(ctx):
        catalog = ctx.obj.make_catalog(init=False)
        catalog_alias = catalog.add_alias(name, alias, sync=sync)
        click.echo(f"Added alias {catalog_alias.alias!r} -> {name}")


@cli.command("remove-alias")
@click.argument(
    "aliases", nargs=-1, required=True, shell_complete=_complete_alias_names
)
@click.option("--sync/--no-sync", default=True)
@click.pass_context
def remove_alias(ctx, aliases, sync):
    """Remove one or more aliases."""
    with click_context_catalog(ctx):
        catalog = ctx.obj.make_catalog(init=False)
        alias_map = {ca.alias: ca for ca in catalog.catalog_aliases}
        with catalog.maybe_synchronizing(sync):
            for alias in aliases:
                if alias not in alias_map:
                    raise click.UsageError(f"Unknown alias: {alias!r}")
                alias_map[alias].remove()
                click.echo(f"Removed alias {alias!r}")


@cli.command("list")
@click.pass_context
def list_entries(ctx):
    """List all entries."""
    with click_context_catalog(ctx):
        catalog = ctx.obj.make_catalog(init=False)
        names = catalog.list() or ("No entries.",)
        for name in names:
            click.echo(name)


@cli.command("list-aliases")
@click.pass_context
def list_aliases(ctx):
    """List all aliases."""
    with click_context_catalog(ctx):
        catalog = ctx.obj.make_catalog(init=False)
        aliases = catalog.list_aliases() or ("No aliases.",)
        for alias in aliases:
            click.echo(alias)


@cli.command()
@click.pass_context
def info(ctx):
    """Show catalog metadata: path, remotes, entry/alias counts."""
    with click_context_catalog(ctx):
        catalog = ctx.obj.make_catalog(init=False)
        click.echo(f"path:    {catalog.repo_path}")
        click.echo(f"commit:  {catalog.repo.head.commit.hexsha[:12]}")
        remotes = tuple(r.name for r in catalog.repo.remotes)
        click.echo(f"remotes: {', '.join(remotes) if remotes else '(none)'}")
        click.echo(f"entries: {len(catalog.list())}")
        click.echo(f"aliases: {len(catalog.list_aliases())}")


@cli.command()
@click.argument("name", shell_complete=_complete_entry_names)
@click.option(
    "-o",
    "--output",
    default=None,
    type=click.Path(),
    help="Output directory (default: current directory).",
)
@click.pass_context
def get(ctx, name, output):
    """Export an entry's tgz to a directory."""
    with click_context_catalog(ctx):
        catalog = ctx.obj.make_catalog(init=False)
        result = catalog.get_tgz(name, dir_path=output)
        click.echo(f"Exported to {result}")


@cli.command()
@click.pass_context
def push(ctx):
    """Push catalog to remote(s)."""
    with click_context_catalog(ctx):
        catalog = ctx.obj.make_catalog(init=False)
        catalog.push()
        click.echo("Pushed.")


@cli.command()
@click.pass_context
def pull(ctx):
    """Pull catalog from remote(s)."""
    with click_context_catalog(ctx):
        catalog = ctx.obj.make_catalog(init=False)
        catalog.pull()
        click.echo("Pulled.")


@cli.command()
@click.pass_context
def sync(ctx):
    """Pull then push."""
    with click_context_catalog(ctx):
        catalog = ctx.obj.make_catalog(init=False)
        catalog.sync()
        click.echo("Synced.")


@cli.command()
@click.argument("url")
@click.option(
    "-n", "--name", "dest_name", default=None, help="Destination catalog name."
)
@click.option(
    "-p",
    "--path",
    "dest_path",
    default=None,
    type=click.Path(),
    help="Destination repo path.",
)
def clone(url, dest_name, dest_path):
    """Clone a catalog from a remote URL."""
    from xorq.catalog.catalog import Catalog

    with click_context_default():
        match (dest_path, dest_name):
            case (None, None):
                repo_path = None
            case (_, None):
                repo_path = dest_path
            case (None, _):
                repo_path = Catalog.name_to_repo_path(dest_name)
            case (_, _):
                raise click.UsageError("--name and --path are mutually exclusive.")
        catalog = Catalog.clone_from(url, repo_path)
        click.echo(f"Cloned to {catalog.repo_path}")


@cli.command()
@click.pass_context
def check(ctx):
    """Validate catalog consistency."""
    with click_context_catalog(ctx):
        catalog = ctx.obj.make_catalog(init=False)
        catalog.assert_consistency()
        click.echo("OK")



import contextlib
import functools
import json
import operator
import pathlib
import sys
from pathlib import Path
from typing import Any, Dict

import dask
import toolz
import yaml
from attr import (
    evolve,
    field,
    frozen,
)
from attr.validators import (
    instance_of,
    is_callable,
    optional,
    or_,
)

import xorq
import xorq.common.utils.logging_utils as lu
import xorq.ibis_yaml.translate  #  noqa: F401
import xorq.vendor.ibis as ibis
import xorq.vendor.ibis.expr.types as ir
from xorq.caching import (
    ParquetCache,
    ParquetSnapshotCache,
    ParquetTTLSnapshotCache,
    SnapshotStrategy,
)
from xorq.common.utils.caching_utils import get_xorq_cache_dir
from xorq.common.utils.dask_normalize.dask_normalize_utils import (
    normalize_read_path_md5sum,
)
from xorq.common.utils.defer_utils import normalize_read_path_stat
from xorq.common.utils.graph_utils import (
    find_all_sources,
    opaque_ops,
    replace_nodes,
    walk_nodes,
)
from xorq.common.utils.name_utils import get_uid_prefix
from xorq.common.utils.node_utils import (
    change_read_table_name,
    recreate,
)
from xorq.config import _backend_init
from xorq.expr.api import deferred_read_parquet, read_parquet
from xorq.expr.relations import (
    CachedNode,
    Read,
)
from xorq.ibis_yaml.common import (
    RefEnum,
    Registry,
    RegistryEnum,
    TranslationContext,
    translate_from_yaml,
    translate_to_yaml,
)
from xorq.ibis_yaml.config import config
from xorq.ibis_yaml.sql import generate_sql_plans
from xorq.ibis_yaml.utils import freeze
from xorq.vendor.ibis.backends.profiles import Profile
from xorq.vendor.ibis.common.collections import FrozenOrderedDict
from xorq.vendor.ibis.expr.operations import DatabaseTable, InMemoryTable


try:
    from enum import StrEnum
except ImportError:
    from strenum import StrEnum


class DumpFiles(StrEnum):
    deferred_reads = "deferred_reads.yaml"
    expr = "expr.yaml"
    metadata = "metadata.json"
    profiles = "profiles.yaml"
    sql = "sql.yaml"


class MemtableTypes(StrEnum):
    inmemory = "memtables"
    database_table = "database_tables"


memory_backends = ("pandas", "duckdb", "datafusion", "xorq")
table_like_ops = tuple(o for o in opaque_ops if issubclass(o, DatabaseTable))


class CleanDictYAMLDumper(yaml.SafeDumper):
    def ignore_aliases(self, data):
        return True

    def represent_enum(self, data):
        return self.represent_scalar("tag:yaml.org,2002:str", data.name)

    def represent_frozenordereddict(self, data):
        return self.represent_dict(dict(data))

    def represent_ibis_schema(self, data):
        schema_dict = {name: str(dtype) for name, dtype in zip(data.names, data.types)}
        return self.represent_mapping("tag:yaml.org,2002:map", schema_dict)

    def represent_posix_path(self, data):
        return self.represent_scalar("tag:yaml.org,2002:str", str(data))

    yaml_representer_pairs = (
        (RefEnum, represent_enum),
        (RegistryEnum, represent_enum),
        (FrozenOrderedDict, represent_frozenordereddict),
        (ibis.Schema, represent_ibis_schema),
        (pathlib.PosixPath, represent_posix_path),
    )

    @classmethod
    def add_representers(cls):
        for to_register, representer in cls.yaml_representer_pairs:
            cls.add_representer(to_register, representer)


CleanDictYAMLDumper.add_representers()


@frozen
class ArtifactStore:
    root_path = field(validator=instance_of(Path), converter=Path)

    def __attrs_post_init__(self):
        self.root_path.mkdir(parents=True, exist_ok=True)

    def get_path(self, *parts) -> pathlib.Path:
        return self.root_path.joinpath(*parts)

    def _read(self, read_f, *parts):
        path = self.get_path(*parts)
        with path.open("r") as f:
            return read_f(f)

    def read_yaml(self, *path_parts) -> Dict[str, Any]:
        return self._read(yaml.safe_load, *path_parts)

    def read_json(self, *path_parts) -> Dict[str, Any]:
        return self._read(json.load, *path_parts)

    def read_text(self, *path_parts) -> str:
        return self._read(operator.methodcaller("read"), *path_parts)

    @contextlib.contextmanager
    def _write(self, *path_parts):
        path = self.get_path(*path_parts)
        path.parent.mkdir(parents=True, exist_ok=True)
        with path.open("w") as f:
            yield (path, f)

    def write_yaml(self, data: Dict[str, Any], *path_parts) -> pathlib.Path:
        with self._write(*path_parts) as (path, f):
            yaml.dump(
                data,
                f,
                Dumper=CleanDictYAMLDumper,
                default_flow_style=False,
                sort_keys=False,
            )
        return path

    def write_text(self, content: str, *path_parts) -> pathlib.Path:
        with self._write(*path_parts) as (path, f):
            f.write(content)
        return path

    def write_parquet(self, table, *path_parts) -> pathlib.Path:
        import pyarrow.parquet as pq

        with self._write(*path_parts) as (path, f):
            pq.write_table(table, path)
        return path

    def exists(self, *path_parts) -> bool:
        return self.get_path(*path_parts).exists()

    def save_yaml(self, yaml_dict: Dict[str, Any], filename) -> pathlib.Path:
        return self.write_yaml(yaml_dict, filename)

    def load_yaml(self, filename) -> Dict[str, Any]:
        return self.read_yaml(filename)

    @staticmethod
    def get_expr_hash(expr) -> str:
        with SnapshotStrategy().normalization_context(expr):
            expr_hash = dask.base.tokenize(expr)[: config.hash_length]
            return expr_hash

    @classmethod
    def from_path_and_expr(cls, builds_dir, expr):
        return cls(root_path=builds_dir.joinpath(cls.get_expr_hash(expr)))


class YamlExpressionTranslator:
    @staticmethod
    def to_yaml(expr: ir.Expr, profiles=(), cache_dir=None) -> Dict[str, Any]:
        context = TranslationContext(
            profiles=freeze(dict(profiles)),
            cache_dir=cache_dir,
        )
        with SnapshotStrategy().normalization_context(expr):
            expr_dict = translate_to_yaml(expr, context)
            expr_dict = freeze(
                expr_dict
                | {
                    RefEnum.schema_ref: context.registry.register_schema(expr.schema())
                    if hasattr(expr, "schema")
                    else None,
                }
            )
            return freeze(
                {
                    "definitions": context.definitions,
                    "expression": expr_dict,
                }
            )

    @staticmethod
    def from_yaml(
        yaml_dict: Dict[str, Any],
        profiles=(),
    ) -> ir.Expr:
        context = TranslationContext(
            registry=Registry(**yaml_dict.get("definitions", {})),
            profiles=freeze(dict(profiles)),
        )
        expr_dict = freeze(yaml_dict["expression"])
        return translate_from_yaml(expr_dict, context)


def dehydrate_cons(cons):
    dehydrated = dict(
        sorted(
            (
                profile.hash_name,
                profile.as_dict()
                | {
                    "kwargs_tuple": dict(profile.as_dict()["kwargs_tuple"]),
                },
            )
            for profile in (con._profile for con in cons)
        )
    )
    return dehydrated


def hydrate_cons(hash_to_profile_kwargs):
    def kwargs_to_con(kwargs):
        match dct := dict(kwargs):
            case {"kwargs_tuple": dict()}:
                dct["kwargs_tuple"] = tuple(dct["kwargs_tuple"].items())
            case _:
                dct["kwargs_tuple"] = tuple(map(tuple, dct["kwargs_tuple"]))
        con = Profile(**dct).get_con()
        return con

    profiles = toolz.valmap(
        kwargs_to_con,
        hash_to_profile_kwargs,
    )
    return profiles


def make_read_op(parquet_path, read_kwargs, con=_backend_init()):  # noqa: B008
    op = deferred_read_parquet(parquet_path, con, **read_kwargs).op()
    args = dict(zip(op.__argnames__, op.__args__))
    op = op.__recreate__(args)
    return op


@frozen
class ExprDumper:
    """
    expr: the expr to be built
    builds_dir: root directory where expr builds are stored
    cache_dir: optional directory for parquet cache files
    debug: when True, output SQL files and debug artifacts (sql.yaml, deferred_reads.yaml)
    """

    expr = field(validator=instance_of(ir.Expr))
    builds_dir = field(validator=instance_of(Path), converter=Path, default="./builds")
    cache_dir = field(validator=optional(instance_of(Path)), factory=get_xorq_cache_dir)
    debug = field(validator=instance_of(bool), default=False)
    read_normalize_method = field(
        validator=is_callable(), default=normalize_read_path_stat
    )

    def __attrs_post_init__(self):
        # use normalize_read_path_md5sum for content addressable / mtime-agnostic hash
        object.__setattr__(
            self,
            "expr",
            self._sanitize_generated_names(self.expr, self.read_normalize_method),
        )
        attrname = "cache_dir"
        match value := getattr(self, attrname):
            case None:
                object.__setattr__(self, attrname, get_xorq_cache_dir())
            case Path():
                pass
            case _:
                object.__setattr__(self, attrname, Path(value))

    @property
    @functools.cache
    def artifact_store(self):
        return ArtifactStore.from_path_and_expr(self.builds_dir, self.expr)

    @property
    @functools.cache
    def expr_path(self):
        return self.artifact_store.root_path

    @property
    def expr_hash(self):
        return self.expr_path.name

    @staticmethod
    def _sanitize_generated_names(expr, normalize_method):
        # InMemoryTable and Read are independent leaf nodes: renaming one cannot
        # affect the other, so a single walk collecting both is equivalent to
        # the previous two sequential walks.
        replacements = {}
        for node in walk_nodes((InMemoryTable, Read), expr):
            if isinstance(node, InMemoryTable):
                if prefix := get_uid_prefix(node.name):
                    name = f"{prefix}{dask.base.tokenize(recreate(node, name='name').to_expr())}"
                    replacements[node] = recreate(
                        node, name=name, normalize_method=normalize_method
                    )
            else:
                if prefix := get_uid_prefix(node.name):
                    table_name = f"{prefix}{dask.base.tokenize(recreate(node, name='name', normalize_method=normalize_method).to_expr())}"
                    replacements[node] = recreate(
                        change_read_table_name(node, table_name=table_name),
                        normalize_method=normalize_method,
                    )
        op = expr.op()
        if replacements:
            op = replace_nodes(replace_from_mapping(replacements), op)
        return op.to_expr()

    def _prepare_expr_file(self, expr, profiles):
        path = self.artifact_store.get_path(DumpFiles.expr)
        # we can't translate to yaml until the memtable parquets are written: they will be tokenized
        writer = toolz.compose(
            functools.partial(self.artifact_store.save_yaml, filename=DumpFiles.expr),
            functools.partial(
                YamlExpressionTranslator.to_yaml,
                expr,
                profiles,
                self.cache_dir,
            ),
        )
        return (path, writer)

    def _prepare_sql_file(self, sql: str) -> str:
        sql_hash = dask.base.tokenize(sql)[: config.hash_length]
        filename = f"{sql_hash}.sql"
        path = self.artifact_store.get_path(filename)
        writer = functools.partial(self.artifact_store.write_text, sql, filename)
        return (path, writer)

    def _prepare_memtable(self, mt, which):
        assert which in MemtableTypes
        table = mt.to_expr().to_pyarrow()
        filename = f"{dask.base.tokenize(table)}.parquet"
        path_parts = (which, filename)
        path = self.artifact_store.get_path(*path_parts)
        writer = functools.partial(
            self.artifact_store.write_parquet,
            table,
            *path_parts,
        )
        return (path, writer)

    def _prepare_sql_plans(
        self,
        sql_plans: Dict[str, Any],
    ) -> Dict[str, Any]:
        queries = {}
        path_to_writer = {}
        for query_name, query_info in sql_plans["queries"].items():
            path, writer = self._prepare_sql_file(query_info["sql"])
            path_to_writer[path] = writer
            queries[query_name] = toolz.dissoc(query_info, "sql") | {
                "sql_file": path.name
            }
        updated_plans = {"queries": queries}
        path_to_writer[self.artifact_store.get_path(DumpFiles.sql)] = functools.partial(
            self.artifact_store.save_yaml,
            updated_plans,
            filename=DumpFiles.sql,
        )
        return path_to_writer

    def _prepare_deferred_reads(
        self,
        deferred_reads: Dict[str, Any],
    ) -> Dict[str, Any]:
        reads = {}
        path_to_writer = {}
        for read_name, read_info in deferred_reads["reads"].items():
            path, writer = self._prepare_sql_file(read_info["sql"])
            path_to_writer[path] = writer
            reads[read_name] = toolz.dissoc(read_info, "sql") | {"sql_file": path.name}
        updated_reads = {"reads": reads}
        path_to_writer[self.artifact_store.get_path(DumpFiles.deferred_reads)] = (
            functools.partial(
                self.artifact_store.save_yaml,
                updated_reads,
                filename=DumpFiles.deferred_reads,
            )
        )
        return path_to_writer

    @staticmethod
    def _make_metadata() -> str:
        metadata = {
            "current_library_version": xorq.__version__,
            "metadata_version": "0.0.0",  # TODO: make it a real thing
            "git_state": lu.get_git_state(hash_diffs=False)
            if lu._git_is_present()
            else None,
            "sys-version_info": tuple(sys.version_info),
        }
        metadata_json = json.dumps(metadata, indent=2)
        return metadata_json

    def _prepare_metadata_file(self):
        path = self.artifact_store.get_path(DumpFiles.metadata)
        writer = functools.partial(
            self.artifact_store.write_text,
            self._make_metadata(),
            DumpFiles.metadata,
        )
        return path, writer

    def _prepare_profiles_file(self, profiles):
        path = self.artifact_store.get_path(DumpFiles.profiles)
        writer = functools.partial(
            self.artifact_store.save_yaml,
            profiles,
            DumpFiles.profiles,
        )
        return path, writer

    def _prepare_debug_info(self):
        sql_plans, deferred_reads = generate_sql_plans(self.expr)
        path_to_writer0 = self._prepare_sql_plans(sql_plans)
        path_to_writer1 = self._prepare_deferred_reads(deferred_reads)
        path_to_writer = path_to_writer0 | path_to_writer1
        return path_to_writer

    def _replace_tables(self, expr):
        """Single-pass replacement of InMemoryTable and qualifying DatabaseTable nodes.

        Combines what were previously two separate walk_nodes + replace_nodes
        calls (_memtables_to_deferred_reads and _replace_inmemory_backend_tables)
        into one graph traversal and one replacement pass.
        """
        path_to_writer = {}
        replacements = {}
        for node in walk_nodes((InMemoryTable, DatabaseTable), expr):
            if isinstance(node, InMemoryTable):
                path, writer = self._prepare_memtable(node, MemtableTypes.inmemory)
                dr_op = make_read_op(
                    parquet_path=path,
                    read_kwargs={
                        "table_name": node.name,
                        "schema": node.schema,
                        str(MemtableTypes.inmemory): True,
                        "normalize_method": self.read_normalize_method,
                    },
                )
            elif (
                isinstance(node, table_like_ops)
                or node.source.name not in memory_backends
            ):
                continue
            else:
                path, writer = self._prepare_memtable(
                    node, MemtableTypes.database_table
                )
                dr_op = make_read_op(
                    parquet_path=path,
                    read_kwargs={
                        "table_name": node.name,
                        # we normalize based on content so we can reproducible hash
                        "normalize_method": normalize_read_path_md5sum,
                        "schema": node.schema,
                    },
                    con=node.source,
                )
            path_to_writer[path] = writer
            replacements[node] = dr_op
        op = expr.op()
        if replacements:
            op = replace_nodes(replace_from_mapping(replacements), op)
        return op.to_expr(), path_to_writer

    def dump_expr(self) -> str:
        # we will mutate the expr below
        expr = self.expr

        # write in-memory data to build dir (single walk + single replacement pass)
        expr, path_to_writer0 = self._replace_tables(expr)

        profiles = dehydrate_cons(find_all_sources(expr))
        path_to_writer2 = dict(
            (
                self._prepare_metadata_file(),
                self._prepare_profiles_file(profiles),
            )
        )
        path_to_writer = path_to_writer0 | path_to_writer2
        if self.debug:
            # write SQL plan and deferred-read artifacts if debug enabled
            path_to_writer |= self._prepare_debug_info()
        for writer in path_to_writer.values():
            writer()

        # expr must be written last
        _, writer = self._prepare_expr_file(expr, profiles)
        writer()
        return self.expr_path


@frozen
class ExprLoader:
    expr_path = field(validator=instance_of(Path), converter=Path)
    cache_dir = field(
        validator=optional(or_(instance_of(Path), instance_of(str))), default=None
    )

    @property
    def expr_hash(self):
        return self.expr_path.name

    @property
    @functools.cache
    def artifact_store(self):
        return ArtifactStore(self.expr_path)

    def load_expr(self):
        profiles = hydrate_cons(self.artifact_store.load_yaml(DumpFiles.profiles))
        yaml_dict = self.artifact_store.load_yaml(DumpFiles.expr)
        expr = YamlExpressionTranslator.from_yaml(yaml_dict, profiles=profiles)
        expr = self.deferred_reads_to_memtables(expr, self.expr_path)
        if self.cache_dir:
            expr = self.replace_base_path(expr, base_path=Path(self.cache_dir))
        return expr

    @staticmethod
    def deferred_reads_to_memtables(loaded, expr_path):
        def deferred_read_to_memtable(dr):
            assert any(key == MemtableTypes.inmemory for key, _ in dr.read_kwargs)
            path = next(v for k, v in dr.read_kwargs if k == "path")
            df = read_parquet(expr_path.joinpath(path)).execute()
            mt = ibis.memtable(df, schema=dr.schema, name=dr.name)
            return mt.op()

        drs = tuple(
            dr
            for dr in walk_nodes(Read, loaded)
            if MemtableTypes.inmemory in dict(dr.read_kwargs)
        )
        replacements = {dr: deferred_read_to_memtable(dr) for dr in drs}
        op = loaded.op()
        if replacements:
            op = replace_nodes(replace_from_mapping(replacements), op)
        return op.to_expr()

    @staticmethod
    def replace_base_path(expr, base_path):
        def replace(node, kwargs):
            if isinstance(node, CachedNode) and isinstance(
                node.cache,
                (ParquetCache, ParquetSnapshotCache, ParquetTTLSnapshotCache),
            ):
                evolved = evolve(
                    node.cache,
                    storage=evolve(
                        node.cache.storage,
                        base_path=base_path,
                    ),
                )
                return node.__recreate__(
                    dict(zip(node.argnames, node.args)) | {"cache": evolved}
                )
            elif kwargs:
                return node.__recreate__(kwargs)
            else:
                return node

        return expr.op().replace(replace).to_expr()


@functools.wraps(ExprLoader)
def load_expr(expr_path, **kwargs):
    expr_loader = ExprLoader(expr_path, **kwargs)
    expr = expr_loader.load_expr()
    return expr


# todo: rename to dump_expr
@functools.wraps(ExprDumper)
def build_expr(expr, **kwargs):
    expr_dumper = ExprDumper(expr, **kwargs)
    expr_path = expr_dumper.dump_expr()
    return expr_path


@toolz.curry
def replace_from_to(from_, to_, node, kwargs):
    if node == from_:
        return to_
    elif kwargs:
        return node.__recreate__(kwargs)
    else:
        return node


@toolz.curry
def replace_from_mapping(mapping, node, kwargs):
    if node in mapping:
        return mapping[node]
    elif kwargs:
        return node.__recreate__(kwargs)
    else:
        return node



import pyarrow as pa
import pytest

import xorq.api as xo
from xorq.common.utils.ibis_utils import from_ibis
from xorq.tests.util import assert_frame_equal


ibis = pytest.importorskip("ibis")


@pytest.fixture(scope="function")
def con():
    return ibis.duckdb.connect()


def test_basic_ops():
    t = ibis.memtable({"id": [1, 2, 3]})
    xorq_t = from_ibis(t)
    assert xorq_t is not None


def test_pyarrow_table_proxy():
    table = pa.table({"id": [1, 2, 3, 4, 5], "value": [10.0, 20.0, 30.0, 40.0, 50.0]})

    t = ibis.memtable(table)
    expr = t.filter(t.id > 2)

    xorq_expr = from_ibis(expr)

    assert_frame_equal(expr.execute(), xorq_expr.execute())


@pytest.mark.parametrize(
    "field,predicate", (("id", ibis._.id == 1), ("name", ibis._.name == "Bob"))
)
def test_select(field, predicate, con):
    t = ibis.memtable({"id": [1, 2, 3], "name": ["Alice", "Bob", "Charlie"]})
    expr = t.select(field).filter(predicate)
    xorq_expr = from_ibis(expr)
    assert xorq_expr is not None
    assert not xo.execute(xorq_expr).empty


def test_read_parquet(con, parquet_dir):
    astronauts_parquet = parquet_dir / "astronauts.parquet"
    t = con.read_parquet(astronauts_parquet, table_name="astronauts")
    expr = t.join(t, "id")
    xorq_expr = from_ibis(expr)
    assert xorq_expr is not None


def test_read_csv(con, csv_dir):
    astronauts_csv = csv_dir / "astronauts.csv"
    t = con.read_csv(astronauts_csv, table_name="astronauts")
    expr = t.join(t, "id")
    xorq_expr = from_ibis(expr)
    assert xorq_expr is not None



"""Operations for working with JSON data."""

from __future__ import annotations

from public import public

import xorq.vendor.ibis.expr.datatypes as dt
import xorq.vendor.ibis.expr.rules as rlz
from xorq.vendor.ibis.common.annotations import attribute
from xorq.vendor.ibis.expr.operations import Value


@public
class JSONGetItem(Value):
    """Get a value from a JSON object or array."""

    arg: Value[dt.JSON]
    index: Value[dt.String | dt.Integer]

    dtype = rlz.dtype_like("arg")
    shape = rlz.shape_like("args")


@public
class ToJSONArray(Value):
    """Convert a value to an array of JSON objects."""

    arg: Value[dt.JSON]

    shape = rlz.shape_like("arg")

    @attribute
    def dtype(self) -> dt.DataType:
        return dt.Array(self.arg.dtype)


@public
class ToJSONMap(Value):
    """Convert a value to a map of string to JSON."""

    arg: Value[dt.JSON]

    shape = rlz.shape_like("arg")

    @attribute
    def dtype(self) -> dt.DataType:
        return dt.Map(dt.string, self.arg.dtype)


@public
class UnwrapJSONString(Value):
    """Unwrap a JSON string into an engine-native string."""

    arg: Value[dt.JSON]

    dtype = dt.string
    shape = rlz.shape_like("arg")


@public
class UnwrapJSONInt64(Value):
    """Unwrap a JSON number into an engine-native int64."""

    arg: Value[dt.JSON]

    dtype = dt.int64
    shape = rlz.shape_like("arg")


@public
class UnwrapJSONFloat64(Value):
    """Unwrap a JSON number into an engine-native float64."""

    arg: Value[dt.JSON]

    dtype = dt.float64
    shape = rlz.shape_like("arg")


@public
class UnwrapJSONBoolean(Value):
    """Unwrap a JSON bool into an engine-native bool."""

    arg: Value[dt.JSON]

    dtype = dt.boolean
    shape = rlz.shape_like("arg")



import itertools
import operator
import random
from operator import methodcaller
from pathlib import PosixPath

import numpy as np
import pandas as pd
import pyarrow as pa
import pytest
from pytest import param

import xorq.api as xo
from xorq.caching import ParquetCache, SourceCache
from xorq.common.exceptions import XorqError
from xorq.expr.relations import register_and_transform_remote_tables
from xorq.loader import load_backend
from xorq.tests.util import assert_frame_equal, check_eq
from xorq.vendor import ibis
from xorq.vendor.ibis import _
from xorq.vendor.ibis.expr.types.relations import CACHED_NODE_NAME_PLACEHOLDER


expected_tables = (
    "array_types",
    "astronauts",
    "awards_players",
    "awards_players_special_types",
    "batting",
    "diamonds",
    "functional_alltypes",
    "geo",
    "geography_columns",
    "geometry_columns",
    "json_t",
    "map",
    "spatial_ref_sys",
    "topk",
    "tzone",
)

KEY_PREFIX = xo.config.options.cache.key_prefix


@pytest.fixture(scope="function")
def ls_con():
    return xo.connect()


@pytest.fixture(scope="session")
def pg():
    conn = xo.postgres.connect_env()
    yield conn
    remove_unexpected_tables(conn)


@pytest.fixture(scope="session")
def alltypes(pg):
    return pg.table("functional_alltypes")


@pytest.fixture(scope="session")
def alltypes_df(alltypes):
    return alltypes.execute()


@pytest.fixture(scope="session")
def trino_table():
    return (
        xo.trino.connect(database="tpch", schema="sf1")
        .table("orders")
        .cast({"orderdate": "date"})
    )


@pytest.fixture(scope="function")
def duckdb_con(csv_dir):
    con = xo.duckdb.connect()
    con.read_csv(csv_dir / "awards_players.csv", table_name="ddb_players")
    con.read_csv(csv_dir / "batting.csv", table_name="batting")
    return con


@pytest.fixture(scope="function")
def pg_batting(pg):
    return pg.table("batting")


@pytest.fixture(scope="session")
def parquet_batting(parquet_dir):
    return parquet_dir / "batting.parquet"


@pytest.fixture(scope="session")
def ls_batting(parquet_batting):
    return xo.connect().read_parquet(parquet_batting)


@pytest.fixture(scope="function")
def ddb_batting(duckdb_con):
    return duckdb_con.create_table(
        "db-batting",
        duckdb_con.table("batting").to_pyarrow(),
    )


@pytest.fixture(scope="function")
def pd_batting(parquet_batting):
    return xo.deferred_read_parquet(
        parquet_batting,
        xo.pandas.connect(),
    )


def make_merged(expr):
    agg = expr.group_by(["custkey", "orderdate"]).agg(
        _.totalprice.sum().name("totalprice")
    )
    w = ibis.window(group_by="custkey", order_by="orderdate")
    windowed = (
        agg.mutate(_.totalprice.cumsum().over(w).name("curev"))
        .mutate(_.curev.lag(1).over(w).name("curev@t-1"))
        .select(["custkey", "orderdate", "curev", "curev@t-1"])
    )
    merged = expr.asof_join(
        windowed,
        on="orderdate",
        predicates="custkey",
    ).select(
        [expr[c] for c in expr.columns]
        + [windowed[c] for c in windowed.columns if c not in expr.columns]
    )
    return merged


def remove_unexpected_tables(dirty):
    # drop tables
    for table in dirty.list_tables():
        if table not in expected_tables:
            dirty.drop_table(table, force=True)

    # drop view
    for table in dirty.list_tables():
        if table not in expected_tables:
            dirty.drop_view(table, force=True)

    if sorted(dirty.list_tables()) != sorted(expected_tables):
        raise ValueError(
            f"unexpected tables after cleanup: {set(dirty.list_tables()) - set(expected_tables)}"
        )


@pytest.mark.postgres
def test_multiple_record_batches(pg):
    con = xo.connect()

    table = pg.table("batting")
    left = con.register(table.to_pyarrow_batches(), "batting_0")
    right = con.register(table.to_pyarrow_batches(), "batting_1")

    expr = (
        left.join(right, "playerID")
        .limit(15)
        .select(player_id="playerID", year_id="yearID_right")
        .cache(SourceCache.from_kwargs(source=con))
    )

    res = expr.execute()
    assert isinstance(res, pd.DataFrame)
    assert 0 < len(res) <= 15


@pytest.mark.postgres
@pytest.mark.parametrize("method", [xo.to_pyarrow, xo.to_pyarrow_batches, xo.execute])
def test_into_backend_simple(pg, method):
    con = xo.connect()
    expr = pg.table("batting").into_backend(con, "ls_batting")
    res = method(expr)

    if isinstance(res, pa.RecordBatchReader):
        res = next(res)

    assert len(res) > 0


@pytest.mark.postgres
@pytest.mark.parametrize("method", ["to_pyarrow", "to_pyarrow_batches", "execute"])
def test_into_backend_complex(pg, method):
    con = xo.connect()

    t = pg.table("batting").into_backend(con, "ls_batting")

    expr = (
        t.join(t, "playerID")
        .limit(15)
        .select(player_id="playerID", year_id="yearID_right")
        .cache(SourceCache.from_kwargs(source=con))
    )

    assert xo.to_sql(expr).count(CACHED_NODE_NAME_PLACEHOLDER) > 0
    res = methodcaller(method)(expr)

    if isinstance(res, pa.RecordBatchReader):
        res = next(res)

    assert 0 < len(res) <= 15


@pytest.mark.postgres
def test_double_into_backend_batches(pg):
    con = xo.connect()
    ddb_con = xo.duckdb.connect()

    t = pg.table("batting").into_backend(con, "ls_batting")

    expr = (
        t.join(t, "playerID")
        .limit(15)
        .into_backend(ddb_con)
        .select(player_id="playerID", year_id="yearID_right")
        .cache(SourceCache.from_kwargs(source=con))
    )

    res = expr.to_pyarrow_batches()
    res = next(res)

    assert len(res) == 15


@pytest.mark.benchmark
def test_into_backend_cache(pg, tmp_path):
    con = xo.connect()
    ddb_con = xo.duckdb.connect()

    t = pg.table("batting").into_backend(con, "ls_batting")

    expr = (
        t.join(t, "playerID")
        .limit(15)
        .cache(SourceCache.from_kwargs(source=con))
        .into_backend(ddb_con)
        .select(player_id="playerID", year_id="yearID_right")
        .cache(ParquetCache.from_kwargs(source=ddb_con, relative_path=tmp_path))
    )

    res = expr.execute()
    assert 0 < len(res) <= 15


@pytest.mark.postgres
def test_into_backend_duckdb(pg):
    ddb = xo.duckdb.connect()
    t = pg.table("batting").into_backend(ddb, "ls_batting")
    expr = (
        t.join(t, "playerID")
        .limit(15)
        .select(player_id="playerID", year_id="yearID_right")
    )

    expr, created = register_and_transform_remote_tables(expr)
    query = ibis.to_sql(expr, dialect="duckdb")

    res = ddb.con.sql(query).df()

    assert query.count("ls_batting") == 2
    assert 0 < len(res) <= 15
    assert len(created) == 3


@pytest.mark.postgres
def test_into_backend_duckdb_expr(pg):
    ddb = xo.duckdb.connect()
    t = pg.table("batting").into_backend(ddb, "ls_batting")
    expr = t.join(t, "playerID").limit(15).select(_.playerID * 2)

    expr, created = register_and_transform_remote_tables(expr)
    query = ibis.to_sql(expr, dialect="duckdb")

    res = ddb.con.sql(query).df()

    assert query.count("ls_batting") == 2
    assert 0 < len(res) <= 15
    assert len(created) == 3


@pytest.mark.trino
def test_into_backend_duckdb_trino(trino_table):
    db_con = xo.duckdb.connect()
    expr = trino_table.head(10_000).into_backend(db_con).pipe(make_merged)

    expr, created = register_and_transform_remote_tables(expr)
    query = ibis.to_sql(expr, dialect="duckdb")

    df = db_con.con.sql(query).df()  # to bypass execute hotfix

    assert isinstance(df, pd.DataFrame)
    assert len(df) > 0
    assert len(created) == 3


@pytest.mark.trino
def test_multiple_into_backend_duckdb_xorq(trino_table):
    db_con = xo.duckdb.connect()
    ls_con = xo.connect()

    expr = (
        trino_table.head(10_000)
        .into_backend(db_con)
        .pipe(make_merged)
        .into_backend(ls_con)[lambda t: t.orderstatus == "F"]
    )

    expr, created = register_and_transform_remote_tables(expr)

    df = expr.execute()

    assert isinstance(df, pd.DataFrame)
    assert len(df) > 0
    assert len(created) == 2


@pytest.mark.benchmark
def test_into_backend_duckdb_trino_cached(trino_table, tmp_path):
    db_con = xo.duckdb.connect()
    expr = (
        trino_table.head(10_000)
        .into_backend(db_con)
        .pipe(make_merged)
        .cache(ParquetCache.from_kwargs(relative_path=tmp_path))
    )
    df = expr.execute()
    assert isinstance(df, pd.DataFrame)
    assert len(df) > 0


@pytest.mark.trino
def test_into_backend_to_pyarrow_batches(trino_table):
    db_con = xo.duckdb.connect()
    df = (
        trino_table.head(10_000)
        .into_backend(db_con)
        .pipe(make_merged)
        .to_pyarrow_batches()
        .read_pandas()
    )
    assert not df.empty


@pytest.mark.postgres
def test_to_pyarrow_batches_simple(pg):
    con = xo.duckdb.connect()

    t = pg.table("batting").into_backend(con, "ls_batting")

    expr = (
        t.join(t, "playerID")
        .limit(15)
        .select(player_id="playerID", year_id="yearID_right")
    )

    df = expr.to_pyarrow_batches().read_pandas()
    assert not df.empty


@pytest.mark.postgres
def test_join(ls_con, alltypes, alltypes_df):
    first_10 = alltypes_df.head(10)
    in_memory = ls_con.create_table("in_memory", first_10).into_backend(
        alltypes.op().source
    )
    expr = alltypes.join(in_memory, predicates=[alltypes.id == in_memory.id])
    actual = expr.execute().sort_values("id")
    expected = pd.merge(
        alltypes_df, first_10, how="inner", on="id", suffixes=("", "_right")
    ).sort_values("id")

    assert_frame_equal(actual, expected)


@pytest.fixture
def union_subsets(alltypes, alltypes_df):
    cols_a, cols_b, cols_c = (alltypes.columns.copy() for _ in range(3))

    random.seed(89)
    random.shuffle(cols_a)
    random.shuffle(cols_b)
    random.shuffle(cols_c)
    assert cols_a != cols_b != cols_c

    cols_a = [ca for ca in cols_a if ca != "timestamp_col"]
    cols_b = [cb for cb in cols_b if cb != "timestamp_col"]
    cols_c = [cc for cc in cols_c if cc != "timestamp_col"]

    a = alltypes.filter((alltypes.id >= 5200) & (alltypes.id <= 5210))[cols_a]
    b = alltypes.filter((alltypes.id >= 5205) & (alltypes.id <= 5215))[cols_b]
    c = alltypes.filter((alltypes.id >= 5213) & (alltypes.id <= 5220))[cols_c]

    da = alltypes_df[(alltypes_df.id >= 5200) & (alltypes_df.id <= 5210)][cols_a]
    db = alltypes_df[(alltypes_df.id >= 5205) & (alltypes_df.id <= 5215)][cols_b]
    dc = alltypes_df[(alltypes_df.id >= 5213) & (alltypes_df.id <= 5220)][cols_c]

    return (a, b, c), (da, db, dc)


@pytest.mark.postgres
@pytest.mark.parametrize("distinct", [False, True], ids=["all", "distinct"])
def test_union(ls_con, union_subsets, distinct):
    (a, _, _), (da, db, dc) = union_subsets

    b = ls_con.create_table("b", db)
    expr = ibis.union(a.into_backend(ls_con), b, distinct=distinct).order_by("id")
    result = expr.execute()

    expected = pd.concat([da, db], axis=0).sort_values("id").reset_index(drop=True)

    if distinct:
        expected = expected.drop_duplicates("id")

    assert_frame_equal(result, expected)


@pytest.mark.postgres
def test_union_mixed_distinct(ls_con, union_subsets):
    (a, _, _), (da, db, dc) = union_subsets

    a = a.into_backend(ls_con)
    b = ls_con.create_table("b", db)
    c = ls_con.create_table("c", dc)

    expr = a.union(b, distinct=True).union(c, distinct=False).order_by("id")
    result = expr.execute()
    expected = pd.concat(
        [pd.concat([da, db], axis=0).drop_duplicates("id"), dc], axis=0
    ).sort_values("id")

    assert_frame_equal(result, expected)


@pytest.mark.parametrize(
    "left_filter",
    [
        lambda t: t.yearID == 2015,
        lambda t: t.yearID != 2015,
        lambda t: t.yearID >= 2015,
        lambda t: t.yearID >= 2014.5,
        lambda t: t.yearID <= 2015,
        lambda t: t.yearID > 2015,
        lambda t: t.yearID < 2015,
        lambda t: ~(t.yearID < 2015),
        lambda t: t.yearID.notnull(),
        lambda t: t.yearID.isnull(),
        lambda t: t.playerID == "wilsobo02",
        lambda t: t.yearID.isin([2015, 2014]),
        lambda t: t.yearID.notin([2015, 2014]),
        lambda t: t.yearID.between(2013, 2016),
    ],
)
def test_join_non_trivial_filters(ls_batting, duckdb_con, left_filter):
    awards_players = duckdb_con.table("ddb_players")
    batting = ls_batting.into_backend(duckdb_con)

    left = batting[left_filter]
    right = awards_players[awards_players.lgID == "NL"].drop("yearID", "lgID")
    right_df = right.execute()
    left_df = left.execute()
    predicate = "playerID"
    result_order = ["playerID", "yearID", "lgID", "stint"]

    expr = left.join(right, predicate, how="inner")
    result = (
        expr.execute()
        .fillna(np.nan)[left.columns]
        .sort_values(result_order)
        .reset_index(drop=True)
    )

    expected = (
        check_eq(
            left_df,
            right_df,
            how="inner",
            on=predicate,
            suffixes=("_x", "_y"),
        )[left.columns]
        .sort_values(result_order)
        .reset_index(drop=True)
    )

    assert_frame_equal(result, expected, check_like=True)


@pytest.mark.parametrize(
    ("predicate", "pandas_value"),
    [
        # True
        param(True, True, id="true"),
        param(ibis.literal(True), True, id="true-literal"),
        param([True], True, id="true-list"),
        param([ibis.literal(True)], True, id="true-literal-list"),
        # only True
        param([True, True], True, id="true-true-list"),
        param(
            [ibis.literal(True), ibis.literal(True)], True, id="true-true-literal-list"
        ),
        param([True, ibis.literal(True)], True, id="true-true-const-expr-list"),
        param([ibis.literal(True), True], True, id="true-true-expr-const-list"),
        # False
        param(False, False, id="false"),
        param(ibis.literal(False), False, id="false-literal"),
        param([False], False, id="false-list"),
        param([ibis.literal(False)], False, id="false-literal-list"),
        # only False
        param([False, False], False, id="false-false-list"),
        param(
            [ibis.literal(False), ibis.literal(False)],
            False,
            id="false-false-literal-list",
        ),
        param([False, ibis.literal(False)], False, id="false-false-const-expr-list"),
        param([ibis.literal(False), False], False, id="false-false-expr-const-list"),
    ],
)
@pytest.mark.parametrize(
    "how",
    [
        "inner",
        "left",
        "right",
        "outer",
    ],
)
def test_join_with_trivial_predicate(
    duckdb_con, awards_players, predicate, how, pandas_value
):
    ddb_players = duckdb_con.table("ddb_players")

    n = 5

    base = awards_players.into_backend(duckdb_con).limit(n)
    ddb_base = ddb_players.limit(n)

    left = base.select(left_key="playerID")
    right = ddb_base.select(right_key="playerID")

    left_df = pd.DataFrame({"key": [True] * n})
    right_df = pd.DataFrame({"key": [pandas_value] * n})

    expected = pd.merge(left_df, right_df, on="key", how=how)

    expr = left.join(right, predicate, how=how)
    result = expr.execute()

    assert len(result) == len(expected)


@pytest.mark.postgres
@pytest.mark.parametrize(
    "backend_name",
    [
        "",
        "duckdb",
    ],
)
def test_multiple_pipes(pg, backend_name):
    """This test address the issue reported on bug #69
    link: https://github.com/letsql/letsql/issues/69

    NOTE
    The previous tests didn't catch it because the con object registered the table batting.
    In this test (and the rest) ls_con is a clean (no tables) letsql connection
    """

    new_con = load_backend(backend_name).connect() if backend_name else xo.connect()
    table_name = "batting"
    pg_t = pg.table(table_name)[lambda t: t.yearID == 2015]
    db_t = new_con.create_table(f"db-{table_name}", pg_t.to_pyarrow())[
        lambda t: t.yearID == 2014
    ]

    expr = pg_t.join(
        db_t.into_backend(pg),
        "playerID",
    )

    assert expr.execute() is not None


@pytest.mark.postgres
@pytest.mark.parametrize(
    "function",
    ["to_pyarrow", "execute", "to_pyarrow_batches"],
)
@pytest.mark.parametrize("remote", [True, False])
def test_duckdb_datafusion_roundtrip(ls_con, pg, duckdb_con, function, remote):
    source = pg if remote else ls_con
    cache = SourceCache.from_kwargs(source=source)

    table_name = "batting"
    pg_t = pg.table(table_name)[lambda t: t.yearID == 2015].cache(cache)

    db_t = duckdb_con.create_table(f"ls-{table_name}", xo.to_pyarrow(pg_t))[
        lambda t: t.yearID == 2014
    ]

    expr = pg_t.join(
        db_t.into_backend(pg if remote else ls_con),
        "playerID",
    )

    assert methodcaller(function)(expr) is not None
    assert any(table_name.startswith(KEY_PREFIX) for table_name in source.list_tables())


@pytest.mark.postgres
@pytest.mark.parametrize(
    "tables",
    [
        param(pair, id="-".join(pair))
        for pair in itertools.combinations_with_replacement(
            ["batting", "pg_batting", "parquet_batting", "ls_batting", "ddb_batting"],
            r=2,
        )
    ],
)
def test_execution_expr_multiple_tables(ls_con, tables, request, mocker):
    left, right = map(request.getfixturevalue, tables)

    left_t = (left if not isinstance(left, PosixPath) else ls_con.read_parquet(left))[
        lambda t: t.yearID == 2015
    ]
    left_backend = left_t._find_backend(use_default=False)

    right_t = (
        right if not isinstance(right, PosixPath) else ls_con.read_parquet(right)
    )[lambda t: t.yearID == 2014]
    right_backend = right_t._find_backend(use_default=False)

    # FIXME there seems to be an issue when doing into_backend from duckdb into duckdb
    expr = left_t.join(
        right_t.into_backend(left_backend)
        if right_backend is not left_backend
        else right_t,
        "playerID",
    )

    native_backend = left_backend is right_backend and left_backend.name != "xorq"
    spy = mocker.spy(left_backend, "to_pyarrow_batches") if native_backend else None

    assert expr.execute() is not None
    assert getattr(spy, "call_count", 0) == int(native_backend)


@pytest.mark.postgres
@pytest.mark.parametrize(
    "tables",
    [
        param(
            pair,
            id="-".join(pair),
        )
        for pair in itertools.permutations(
            ["pg_batting", "ls_batting", "pd_batting", "ddb_batting"],
            r=2,
        )
    ],
)
def test_execution_expr_multiple_tables_cached(ls_con, tables, request):
    from xorq.caching import ParquetCache, SourceCache

    table_name = "batting"
    left, right = map(request.getfixturevalue, tables)
    (left_source, right_source) = (expr.op().source for expr in (left, right))
    (left_cache, right_cache) = (
        SourceCache.from_kwargs(source=source)
        if source.name != "duckdb"
        else ParquetCache.from_kwargs(source=ls_con)
        for source in (left_source, right_source)
    )
    (left_filter, right_filter) = (
        operator.eq(xo._.yearID, year) for year in (2015, 2014)
    )

    left_t = left.into_backend(ls_con, name=f"left-{table_name}").filter(left_filter)
    right_t = right.into_backend(ls_con, name=f"right-{table_name}").filter(
        right_filter
    )
    joined = (
        left_t.cache(left_cache)
        .join(
            right_t.cache(right_cache).into_backend(left_cache.storage.source),
            "playerID",
        )
        .cache(left_cache)
    )

    actual = joined.execute()
    expected = left_t.join(
        right_t.into_backend(left_t._find_backend()),
        "playerID",
    ).execute()
    assert not actual.empty
    columns = list(actual.columns)
    assert_frame_equal(
        actual.sort_values(columns, ignore_index=True),
        expected.sort_values(columns, ignore_index=True),
    )


@pytest.mark.postgres
def test_no_registration_same_table_name(ls_con, pg_batting):
    ddb_con = xo.duckdb.connect()
    table = pg_batting[["playerID", "yearID"]].to_pyarrow()
    ddb_batting = ddb_con.create_table("batting", table)
    ls_batting = ls_con.register(table, "batting")

    expr = ddb_batting.join(
        ls_batting.into_backend(ddb_con),
        "playerID",
    )

    assert expr.execute() is not None


@pytest.mark.postgres
@pytest.mark.parametrize("backend_name", ["", "duckdb"])
def test_multi_engine_cache(pg, ls_con, ls_batting, tmp_path, backend_name):
    other_con = load_backend(backend_name).connect() if backend_name else xo.connect()
    table_name = "batting"
    pg_t = pg.table(table_name)[lambda t: t.yearID > 2014]
    db_t = other_con.create_table(f"db-{table_name}", ls_batting.to_pyarrow())[
        lambda t: t.stint == 1
    ].into_backend(pg)

    expr = pg_t.join(
        db_t,
        db_t.columns,
    ).cache(
        cache=ParquetCache.from_kwargs(
            source=ls_con,
            relative_path=tmp_path,
        )
    )

    assert expr.execute() is not None


def test_multi_backend(parquet_dir):
    batting = xo.deferred_read_parquet(
        parquet_dir.joinpath("batting.parquet"), con=xo.connect()
    )
    awards_players = xo.deferred_read_parquet(
        parquet_dir.joinpath("awards_players.parquet"), con=xo.connect()
    )
    on = set(batting.columns).intersection(awards_players.columns)
    joined = batting.select(on).join(awards_players.select(on), predicates=on)
    with pytest.raises(XorqError, match="Multiple backends"):
        joined.execute()



"""Sort key operations."""

from __future__ import annotations

from public import public

import xorq.vendor.ibis.expr.rules as rlz
from xorq.vendor.ibis.expr.operations.core import Value


# TODO(kszucs): move the content of this file to generic.py


# TODO(kszucs): consider to limit its shape to Columnar, we could treat random()
# as a columnar operation too
@public
class SortKey(Value):
    """A sort key."""

    # TODO(kszucs): rename expr to arg or something else except expr
    expr: Value
    ascending: bool = True
    nulls_first: bool = False

    dtype = rlz.dtype_like("expr")
    shape = rlz.shape_like("expr")

    @classmethod
    def __coerce__(cls, key, T=None, S=None):
        key = super().__coerce__(key, T=T, S=S)

        if isinstance(key, cls):
            return key
        else:
            return cls(key)

    @property
    def name(self) -> str:
        return self.expr.name

    @property
    def descending(self) -> bool:
        return not self.ascending



import random

import pytest
from pytest import param

from xorq.common.utils.ibis_utils import from_ibis


ibis = pytest.importorskip("ibis")


@pytest.fixture
def union_subsets(ibis_alltypes):
    randomizer = random.Random(42)
    cols_a, cols_b, cols_c = (
        randomizer.sample(ibis_alltypes.columns, k=len(ibis_alltypes.columns))
        for _ in range(3)
    )
    assert cols_a != cols_b != cols_c

    a = ibis_alltypes.filter((ibis._.id >= 5200) & (ibis._.id <= 5210))[cols_a]
    b = ibis_alltypes.filter((ibis._.id >= 5205) & (ibis._.id <= 5215))[cols_b]
    c = ibis_alltypes.filter((ibis._.id >= 5213) & (ibis._.id <= 5220))[cols_c]

    return a, b, c


@pytest.mark.parametrize("distinct", [False, True], ids=["all", "distinct"])
def test_union(union_subsets, distinct):
    a, b, c = union_subsets

    expr = ibis.union(a, b, distinct=distinct).order_by("id")

    xorq_expr = from_ibis(expr)
    assert xorq_expr.execute() is not None


def test_union_mixed_distinct(union_subsets):
    a, b, c = union_subsets

    expr = a.union(b, distinct=True).union(c, distinct=False).order_by("id")

    xorq_expr = from_ibis(expr)
    assert xorq_expr.execute() is not None


@pytest.mark.parametrize(
    "distinct",
    [
        param(False, id="all"),
        param(True, id="distinct"),
    ],
)
def test_intersect(ibis_alltypes, distinct):
    a = ibis_alltypes.filter((ibis._.id >= 5200) & (ibis._.id <= 5210))
    b = ibis_alltypes.filter((ibis._.id >= 5205) & (ibis._.id <= 5215))
    c = ibis_alltypes.filter((ibis._.id >= 5195) & (ibis._.id <= 5208))

    expr = ibis.intersect(a, b, c, distinct=distinct).order_by("id")

    xorq_expr = from_ibis(expr)
    assert xorq_expr.execute() is not None


@pytest.mark.parametrize(
    "distinct",
    [
        param(False, id="all"),
        param(True, id="distinct"),
    ],
)
def test_difference(ibis_alltypes, distinct):
    a = ibis_alltypes.filter((ibis._.id >= 5200) & (ibis._.id <= 5210))
    b = ibis_alltypes.filter((ibis._.id >= 5205) & (ibis._.id <= 5215))
    c = ibis_alltypes.filter((ibis._.id >= 5195) & (ibis._.id <= 5202))

    expr = ibis.difference(a, b, c, distinct=distinct).order_by("id")

    xorq_expr = from_ibis(expr)
    assert xorq_expr.execute() is not None



from __future__ import annotations

import contextlib
import operator
from functools import singledispatch

import sqlglot as sg
import sqlglot.expressions as sge
import sqlglot.optimizer as sgo
import sqlglot.planner as sgp
from public import public

import xorq.common.exceptions as com
import xorq.vendor.ibis.expr.datatypes as dt
import xorq.vendor.ibis.expr.schema as sch
import xorq.vendor.ibis.expr.types as ir
from xorq.vendor import ibis
from xorq.vendor.ibis.backends.sql.datatypes import SqlglotType
from xorq.vendor.ibis.util import experimental


class Catalog(dict[str, sch.Schema]):
    """A catalog of tables and their schemas."""

    typemap = {
        dt.Int8: "tinyint",
        dt.Int16: "smallint",
        dt.Int32: "int",
        dt.Int64: "bigint",
        dt.Float16: "halffloat",
        dt.Float32: "float",
        dt.Float64: "double",
        dt.Decimal: "decimal",
        dt.Boolean: "boolean",
        dt.JSON: "json",
        dt.Interval: "interval",
        dt.Timestamp: "datetime",
        dt.Date: "date",
        dt.Binary: "varbinary",
        dt.String: "varchar",
        dt.Array: "array",
        dt.Map: "map",
        dt.UUID: "uuid",
        dt.Struct: "struct",
    }

    def to_sqlglot_dtype(self, dtype: dt.DataType) -> str:
        if dtype.is_geospatial():
            return dtype.geotype
        else:
            default = dtype.__class__.__name__.lower()
            return self.typemap.get(type(dtype), default)

    def to_sqlglot_schema(self, schema: sch.Schema) -> dict[str, str]:
        return {name: self.to_sqlglot_dtype(dtype) for name, dtype in schema.items()}

    def to_sqlglot(self):
        return {
            name: self.to_sqlglot_schema(table.schema()) for name, table in self.items()
        }

    def overlay(self, step):
        updates = {dep.name: convert(dep, catalog=self) for dep in step.dependencies}

        # handle scan aliases: FROM foo AS bar
        source = getattr(step, "source", None)
        alias = getattr(source, "args", {}).get("alias")
        if alias is not None and (source_name := self.get(source.name)) is not None:
            self[alias.name] = source_name

        return Catalog({**self, **updates})


@singledispatch
def convert(step, catalog):
    raise TypeError(type(step))


@convert.register(sgp.Scan)
def convert_scan(scan, catalog):
    catalog = catalog.overlay(scan)

    table = catalog[scan.source.alias_or_name]

    if scan.condition:
        pred = convert(scan.condition, catalog=catalog)
        table = table.filter(pred)

    if scan.projections:
        projs = [convert(proj, catalog=catalog) for proj in scan.projections]
        table = table.select(projs)

    if isinstance(scan.limit, int):
        table = table.limit(scan.limit)

    return table


def qualify_projections(projections, groups):
    # The sqlglot planner will (sometimes) alias projections to the aggregate
    # that precedes it.
    #
    # - Sort: lineitem (132849388268768)
    #   Context:
    #     Key:
    #       - "l_returnflag"
    #       - "l_linestatus"
    #   Projections:
    #     - lineitem._g0 AS "l_returnflag"
    #     - lineitem._g1 AS "l_linestatus"
    #     <snip>
    #   Dependencies:
    #   - Aggregate: lineitem (132849388268864)
    #     Context:
    #       Aggregations:
    #         <snip>
    #       Group:
    #         - "lineitem"."l_returnflag"  <-- this is _g0
    #         - "lineitem"."l_linestatus"  <-- this is _g1
    #         <snip>
    #
    #  These aliases are stored in a dictionary in the aggregate `groups`, so if
    #  those are pulled out beforehand then we can use them to replace the
    #  aliases in the projections.

    def transformer(node):
        if isinstance(node, sge.Alias) and (name := node.this.name).startswith("_g"):
            return groups[name]
        return node

    projects = [project.transform(transformer) for project in projections]

    return projects


@convert.register(sgp.Sort)
def convert_sort(sort, catalog):
    catalog = catalog.overlay(sort)

    table = catalog[sort.name]

    if sort.key:
        keys = [convert(key, catalog=None) for key in sort.key]
        table = table.order_by(keys)

    if sort.projections:
        groups = {}
        # group definitions that may be used in projections are defined
        # in the aggregate in dependencies...
        for dep in sort.dependencies:
            if (group := getattr(dep, "group", None)) is not None:
                groups |= group
        projs = [
            convert(proj, catalog=catalog)
            for proj in qualify_projections(sort.projections, groups)
        ]
        table = table.select(projs)

    if isinstance(sort.limit, int):
        table = table.limit(sort.limit)

    return table


_join_types = {
    "": "inner",
    "LEFT": "left",
    "RIGHT": "right",
}


@convert.register(sgp.Join)
def convert_join(join, catalog):
    catalog = catalog.overlay(join)

    left_name = join.name
    left_table = catalog[left_name]

    for right_name, desc in join.joins.items():
        right_table = catalog[right_name]
        join_kind = _join_types[desc["side"]]

        predicate = None
        if desc["join_key"]:
            for left_key, right_key in zip(desc["source_key"], desc["join_key"]):
                left_key = convert(left_key, catalog=catalog)
                right_key = convert(right_key, catalog=catalog)
                if predicate is None:
                    predicate = left_key == right_key
                else:
                    predicate &= left_key == right_key

        if "condition" in desc.keys():
            condition = desc["condition"]
            if predicate is None:
                predicate = convert(condition, catalog=catalog)
            else:
                predicate &= convert(condition, catalog=catalog)

        left_table = left_table.join(right_table, predicates=predicate, how=join_kind)

    if join.condition:
        predicate = convert(join.condition, catalog=catalog)
        left_table = left_table.filter(predicate)

    catalog[left_name] = left_table

    return left_table


def replace_operands(agg):
    # The sqlglot planner will pull out computed operands into a separate
    # section and alias them #
    # e.g.
    # Context:
    #   Aggregations:
    #     - SUM("_a_0") AS "sum_disc_price"
    #   Operands:
    #     - "lineitem"."l_extendedprice" * (1 - "lineitem"."l_discount") AS _a_0
    #
    # For the purposes of decompiling, we want these to be inline, so here we
    # replace those new aliases with the parsed sqlglot expression
    operands = {operand.alias: operand.this for operand in agg.operands}

    def transformer(node):
        if isinstance(node, sge.Column) and node.name in operands.keys():
            return operands[node.name]
        return node

    aggs = [item.transform(transformer) for item in agg.aggregations]

    agg.aggregations = aggs

    return agg


@convert.register(sgp.Aggregate)
def convert_aggregate(agg, catalog):
    catalog = catalog.overlay(agg)

    agg = replace_operands(agg)

    table = catalog[agg.source]
    if agg.aggregations:
        metrics = [convert(a, catalog=catalog) for a in agg.aggregations]
        groups = [convert(g, catalog=catalog) for k, g in agg.group.items()]
        table = table.aggregate(metrics, by=groups)

    return table


@convert.register(sge.Subquery)
def convert_subquery(subquery, catalog):
    tree = sgo.optimize(subquery.this, catalog.to_sqlglot(), rules=sgo.RULES)
    plan = sgp.Plan(tree)
    return convert(plan.root, catalog=catalog)


@convert.register(sge.Literal)
def convert_literal(literal, catalog):
    value = literal.this
    if literal.is_int:
        value = int(value)
    elif literal.is_number:
        value = float(value)
    return ibis.literal(value)


@convert.register(sge.Boolean)
def convert_boolean(boolean, catalog):
    return ibis.literal(boolean.this)


@convert.register(sge.Alias)
def convert_alias(alias, catalog):
    this = convert(alias.this, catalog=catalog)
    return this.name(alias.alias)


@convert.register(sge.Column)
def convert_column(column, catalog):
    table = catalog[column.table]
    return table[column.name]


@convert.register(sge.Ordered)
def convert_ordered(ordered, catalog):
    this = ibis._[ordered.this.name]
    desc = ordered.args.get("desc", False)  # not exposed as an attribute
    nulls_first = ordered.args.get("nulls_first", False)
    return (
        ibis.desc(this, nulls_first=nulls_first)
        if desc
        else ibis.asc(this, nulls_first=nulls_first)
    )


_unary_operations = {
    sge.Paren: lambda x: x,
}


@convert.register(sge.Unary)
def convert_unary(unary, catalog):
    op = _unary_operations[type(unary)]
    this = convert(unary.this, catalog=catalog)
    return op(this)


_binary_operations = {
    sge.LT: operator.lt,
    sge.LTE: operator.le,
    sge.GT: operator.gt,
    sge.GTE: operator.ge,
    sge.EQ: operator.eq,
    sge.NEQ: operator.ne,
    sge.Add: operator.add,
    sge.Sub: operator.sub,
    sge.Mul: operator.mul,
    sge.Div: operator.truediv,
    sge.Pow: operator.pow,
    sge.And: operator.and_,
    sge.Or: operator.or_,
}


@convert.register(sge.Binary)
def convert_binary(binary, catalog):
    op = _binary_operations[type(binary)]
    this = convert(binary.this, catalog=catalog)
    expr = convert(binary.expression, catalog=catalog)

    if isinstance(binary.expression, sge.Subquery):
        # expr is a table expression
        assert len(expr.columns) == 1
        name = expr.columns[0]
        expr = expr[name]

    return op(this, expr)


_reduction_methods = {
    sge.Max: "max",
    sge.Min: "min",
    sge.Quantile: "quantile",
    sge.Sum: "sum",
    sge.Avg: "mean",
}


@convert.register(sge.AggFunc)
def convert_sum(reduction, catalog):
    method = _reduction_methods[type(reduction)]
    this = convert(reduction.this, catalog=catalog)
    return getattr(this, method)()


@convert.register(sge.In)
def convert_in(in_, catalog):
    this = convert(in_.this, catalog=catalog)
    candidates = [convert(expression, catalog) for expression in in_.expressions]
    return this.isin(candidates)


@convert.register(sge.Cast)
def cast(cast, catalog):
    this = convert(cast.this, catalog)
    to = convert(cast.to, catalog)

    return this.cast(to)


@convert.register(sge.DataType)
def datatype(datatype, catalog):
    return SqlglotType().to_ibis(datatype)


@convert.register(sge.Count)
def count(count, catalog):
    return ibis._.count()


@public
@experimental
def parse_sql(sqlstring, catalog, dialect=None):
    """Parse a SQL string into an Ibis expression.

    Parameters
    ----------
    sqlstring : str
        SQL string to parse
    catalog : dict
        A dictionary mapping table names to either schemas or ibis table expressions.
        If a schema is passed, a table expression will be created using the schema.
    dialect : str, optional
        The SQL dialect to use with sqlglot to parse the query string.

    Returns
    -------
    expr : ir.Expr

    """
    catalog = Catalog(
        {name: ibis.table(schema, name=name) for name, schema in catalog.items()}
    )

    expr = sg.parse_one(sqlstring, dialect)
    tree = sgo.optimize(expr, catalog.to_sqlglot(), rules=sgo.RULES)
    plan = sgp.Plan(tree)

    return convert(plan.root, catalog=catalog)


class SQLString(str):
    """Object to hold a formatted SQL string.

    Syntax highlights in Jupyter notebooks.
    """

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({str(self)!r})"

    def _repr_markdown_(self) -> str:
        return f"```sql\n{self!s}\n```"

    def _repr_pretty_(self, p, cycle) -> str:
        output = str(self)
        try:
            from pygments import highlight
            from pygments.formatters import TerminalFormatter
            from pygments.lexers import SqlLexer
        except ImportError:
            pass
        else:
            with contextlib.suppress(Exception):
                output = highlight(
                    code=output,
                    lexer=SqlLexer(),
                    formatter=TerminalFormatter(),
                )

        # strip trailing newline
        p.text(output.strip())


@public
def to_sql(
    expr: ir.Expr, dialect: str | None = None, pretty: bool = True, **kwargs
) -> SQLString:
    """Return the formatted SQL string for an expression.

    Parameters
    ----------
    expr
        Ibis expression.
    dialect
        SQL dialect to use for compilation.
    pretty
        Whether to use pretty formatting.
    kwargs
        Scalar parameters

    Returns
    -------
    str
        Formatted SQL string

    """
    import xorq.vendor.ibis.backends.sql.compilers as sc

    # try to infer from a non-str expression or if not possible fallback to
    # the default pretty dialect for expressions
    if dialect is None:
        try:
            compiler_provider = expr._find_backend(use_default=True)
        except com.XorqError:
            # default to duckdb for SQL compilation because it supports the
            # widest array of ibis features for SQL backends
            compiler_provider = sc.duckdb
    else:
        try:
            compiler_provider = getattr(sc, dialect)
        except AttributeError as e:
            raise ValueError(f"Unknown dialect {dialect}") from e

    if (compiler := getattr(compiler_provider, "compiler", None)) is None:
        raise NotImplementedError(f"{compiler_provider} is not a SQL backend")

    out = compiler.to_sqlglot(expr.unbind(), **kwargs)
    queries = out if isinstance(out, list) else [out]
    dialect = compiler.dialect
    sql = ";\n".join(query.sql(dialect=dialect, pretty=pretty) for query in queries)
    return SQLString(sql)



from typing import Optional

import xorq.vendor.ibis.expr.operations as ops


class PyIcebergTable(ops.DatabaseTable):
    """A table that is bound to a PyIceberg backend."""

    snapshot_id: Optional[int] = None



import pytest
from pytest import param

import xorq.api as xo


@pytest.mark.parametrize(
    "metrics",
    [
        (xo._.int_col.first(),),
        (xo._.bool_col.any(),),
        (xo._.int_col.first(), xo._.bool_col.any()),
    ],
)
def test_first(compiler, alltypes, metrics):
    expr = alltypes.group_by(xo._.tinyint_col).aggregate(
        *metrics,
    )

    yaml_dict = compiler.to_yaml(expr)
    roundtrip_expr = compiler.from_yaml(yaml_dict)

    assert roundtrip_expr.equals(expr)


@pytest.mark.parametrize(
    "result_fn",
    [
        param(
            lambda t, where: t.bool_col.count(where=where),
            id="count",
        ),
        param(
            lambda t, where: t.bool_col.nunique(where=where),
            id="nunique",
        ),
        param(
            lambda t, where: t.bool_col.any(where=where),
            id="any",
        ),
        param(
            lambda t, where: t.bool_col.notany(where=where),
            id="notany",
        ),
        param(
            lambda t, where: -t.bool_col.any(where=where),
            id="any_negate",
        ),
        param(
            lambda t, where: t.bool_col.all(where=where),
            id="all",
        ),
        param(
            lambda t, where: t.bool_col.notall(where=where),
            id="notall",
        ),
        param(
            lambda t, where: -t.bool_col.all(where=where),
            id="all_negate",
        ),
        param(
            lambda t, where: t.double_col.sum(where=where),
            id="sum",
        ),
        param(
            lambda t, where: (t.int_col > 0).sum(where=where),
            id="bool_sum",
        ),
        param(
            lambda t, where: t.double_col.mean(where=where),
            id="mean",
        ),
        param(
            lambda t, where: t.double_col.min(where=where),
            id="min",
        ),
        param(
            lambda t, where: t.double_col.max(where=where),
            id="max",
        ),
        param(
            lambda t, where: t.double_col.std(how="sample", where=where),
            id="std",
        ),
        param(
            lambda t, where: t.double_col.var(how="sample", where=where),
            id="var",
        ),
        param(
            lambda t, where: t.double_col.std(how="pop", where=where),
            id="std_pop",
        ),
        param(
            lambda t, where: t.double_col.var(how="pop", where=where),
            id="var_pop",
        ),
        param(
            lambda t, where: t.string_col.nunique(where=where),
            id="string_nunique",
        ),
        param(
            lambda t, where: t.double_col.first(where=where),
            id="first",
        ),
        param(
            lambda t, where: t.double_col.last(where=where),
            id="last",
        ),
        param(
            lambda t, where: t.bigint_col.bit_and(where=where),
            id="bit_and",
        ),
        param(
            lambda t, where: t.bigint_col.bit_or(where=where),
            id="bit_or",
        ),
        param(
            lambda t, where: t.bigint_col.bit_xor(where=where),
            id="bit_xor",
        ),
        param(
            lambda t, where: t.count(where=where),
            id="count_star",
        ),
    ],
)
@pytest.mark.parametrize(
    "ibis_cond",
    [
        param(lambda _: None, id="no_cond"),
        param(
            lambda t: t.string_col.isin(["1", "7"]),
            id="is_in",
        ),
        param(
            lambda _: xo._.string_col.isin(["1", "7"]),
            id="is_in_deferred",
        ),
    ],
)
def test_reduction_ops(
    compiler,
    alltypes,
    result_fn,
    ibis_cond,
):
    expr = alltypes.agg(tmp=result_fn(alltypes, ibis_cond(alltypes))).tmp

    yaml_dict = compiler.to_yaml(expr)
    roundtrip_expr = compiler.from_yaml(yaml_dict)

    assert roundtrip_expr.equals(expr)


def test_approx_median(compiler, alltypes):
    expr = alltypes.double_col.approx_median()

    yaml_dict = compiler.to_yaml(expr)
    roundtrip_expr = compiler.from_yaml(yaml_dict)

    assert roundtrip_expr.equals(expr)


def test_median(compiler, alltypes):
    expr = alltypes.double_col.median()

    yaml_dict = compiler.to_yaml(expr)
    roundtrip_expr = compiler.from_yaml(yaml_dict)

    assert roundtrip_expr.equals(expr)



import pytest

import xorq.api as xo


@pytest.mark.parametrize("file_format", ["parquet", "csv"])
def test_sql_on_deferred_read(file_format, request):
    ddb = xo.duckdb.connect()
    diamonds_path = (
        request.getfixturevalue(f"{file_format}_dir") / f"diamonds.{file_format}"
    )

    deferred_read = getattr(xo, f"deferred_read_{file_format}")

    expr = (
        deferred_read(diamonds_path, con=ddb, table_name="diamonds_ddb")
        .limit(2)
        .select(xo._.y, xo._.z, xo._.cut)
        .sql("SELECT cut, avg(y + z) as c FROM diamonds_ddb group by cut")
    )

    assert not expr.execute().empty



from typing import Any, Mapping

import pyarrow as pa

from xorq.vendor.ibis.backends.duckdb import Backend as IbisDuckDBBackend
from xorq.vendor.ibis.expr import types as ir
from xorq.vendor.ibis.util import gen_name


class Backend(IbisDuckDBBackend):
    def execute(
        self,
        expr: ir.Expr,
        params: Mapping | None = None,
        limit: str | None = "default",
        **_: Any,
    ) -> Any:
        batch_reader = self.to_pyarrow_batches(expr, params=params, limit=limit)
        return expr.__pandas_result__(
            batch_reader.read_pandas(timestamp_as_object=True)
        )

    def read_record_batches(self, source, table_name=None):
        table_name = table_name or gen_name("read_record_batches")
        self.con.register(table_name, source)
        return self.table(table_name)

    def to_pyarrow_batches(
        self,
        expr: ir.Expr,
        *,
        params: Mapping[ir.Scalar, Any] | None = None,
        limit: int | str | None = None,
        chunk_size: int = 10_000,
        **_: Any,
    ) -> pa.ipc.RecordBatchReader:
        return self._to_duckdb_relation(
            expr, params=params, limit=limit
        ).fetch_arrow_reader(chunk_size)



import shutil
from pathlib import Path

import pytest

import xorq.api as xo
from xorq.catalog.catalog import (
    BuildTgz,
    Catalog,
    CatalogAddition,
    CatalogAlias,
    CatalogEntry,
    with_pure_suffix,
)
from xorq.catalog.constants import CatalogInfix
from xorq.catalog.expr_utils import (
    build_expr_context_tgz,
)
from xorq.catalog.tar_utils import (
    REQUIRED_TGZ_NAMES,
    write_tgz,
)
from xorq.catalog.tests.conftest import (
    compare_repo_and_catalog,
)


def test_catalog_add(catalog, data_dict):
    catalog_entries = tuple(catalog.add(path) for path in data_dict.values())
    assert all(catalog_entry.exists() for catalog_entry in catalog_entries)
    for catalog_entry in catalog_entries:
        catalog_entry.assert_consistency()
    catalog.assert_consistency()
    assert set(catalog.list()) == {
        with_pure_suffix(path).name for path in data_dict.values()
    }

    # test not exists condition
    path = next(iter(data_dict.values()))
    with pytest.raises(AssertionError):
        catalog.add(path)


def test_catalog_addition_from_expr(catalog):
    expr = xo.memtable({"from-expr": ["from-expr"]})
    catalog_addition = CatalogAddition.from_expr(expr, catalog)
    assert catalog_addition.build_tgz.path.exists()
    assert catalog_addition._maybe_tmpfile is not None
    catalog_entry = catalog_addition.add()
    assert catalog_entry.exists()
    catalog.assert_consistency()
    assert catalog_entry.name in catalog.list()


def test_catalog_addition_with_aliases(catalog):
    from attr import evolve

    expr = xo.memtable({"with-aliases": ["with-aliases"]})
    catalog_addition = CatalogAddition.from_expr(expr, catalog)
    aliases = ("alias-x", "alias-y")
    catalog_addition = evolve(catalog_addition, aliases=aliases)
    catalog_entry = catalog_addition.add()

    commit_message = catalog.repo.head.commit.message.strip()
    assert all(alias in commit_message for alias in aliases)

    assert catalog_entry.exists()
    assert {ca.alias for ca in catalog_entry.aliases} == set(aliases)
    catalog.assert_consistency()


def test_catalog_addition_from_expr_tmpfile_lifecycle(catalog):
    expr = xo.memtable({"lifecycle": ["lifecycle"]})
    catalog_addition = CatalogAddition.from_expr(expr, catalog)
    tgz_path = catalog_addition.build_tgz.path
    assert tgz_path.exists()
    del catalog_addition
    assert not tgz_path.exists()


def test_catalog_rm(catalog, data_dict):
    catalog_entries = tuple(catalog.add(path) for path in data_dict.values())
    for catalog_entry in catalog_entries:
        catalog.remove(catalog_entry.name)
    assert not any(catalog_entry.exists() for catalog_entry in catalog_entries)
    for catalog_entry in catalog_entries:
        catalog_entry.assert_consistency()
    catalog.assert_consistency()
    assert not catalog.list()

    # test exists condition
    name = next(iter(data_dict.keys()))
    with pytest.raises(AssertionError):
        catalog.remove(name)


def test_catalog_rm_removes_aliases(catalog_populated):
    name = catalog_populated.list()[0]
    alias_a = "alias-one"
    alias_b = "alias-two"
    catalog_populated.add_alias(name, alias_a)
    catalog_populated.add_alias(name, alias_b)

    catalog_entry = catalog_populated.get_catalog_entry(name)
    assert len(catalog_entry.aliases) == 2

    catalog_populated.remove(name)

    commit_message = catalog_populated.repo.head.commit.message.strip()
    assert alias_a in commit_message
    assert alias_b in commit_message

    assert not catalog_entry.exists()
    assert not any(
        ca.alias_path.exists()
        for ca in (catalog_populated.catalog_aliases or [])
        if ca.alias in (alias_a, alias_b)
    )
    catalog_populated.assert_consistency()


def test_catalog_clone_from_push(repo_cloned_bare, tmpdir):
    cloned = Catalog.clone_from(
        repo_cloned_bare.working_dir, Path(tmpdir).joinpath("cloned")
    )
    before = cloned.list()
    compare_repo_and_catalog(repo_cloned_bare, cloned)

    with build_expr_context_tgz(xo.memtable({"to-push": ["to-push"]})) as tgz_path:
        cloned.add(tgz_path)
        cloned.push()

    after = cloned.list()
    compare_repo_and_catalog(repo_cloned_bare, cloned)

    assert before != after


@pytest.mark.parametrize("elide", REQUIRED_TGZ_NAMES)
def test_test_tgz(elide, catalog, tmpdir):
    tgz_path = write_tgz(
        Path(tmpdir).joinpath("build.tgz"),
        {name: b"" for name in REQUIRED_TGZ_NAMES if name != elide},
    )
    with pytest.raises(AssertionError, match=elide):
        BuildTgz(tgz_path)


def test_assert_consistency(catalog, tmpdir):
    tgz_path = write_tgz(
        Path(tmpdir).joinpath("build.tgz"),
        dict.fromkeys(REQUIRED_TGZ_NAMES, b""),
    )
    catalog_addition = CatalogAddition(BuildTgz(tgz_path), catalog)
    catalog_addition.ensure_dirs()
    catalog_path = catalog_addition.catalog_entry.catalog_path
    with catalog.commit_context("bad commit"):
        shutil.copy(
            tgz_path,
            catalog_path,
        )
        catalog.repo.index.add((catalog_path,))
    with pytest.raises(AssertionError):
        catalog.assert_consistency()
    with pytest.raises(AssertionError):
        CatalogEntry(catalog_addition.name, catalog, False)


def test_add_alias(catalog_populated):
    name = catalog_populated.list()[0]
    alias = "my-alias"
    catalog_alias = catalog_populated.add_alias(name, alias)

    assert isinstance(catalog_alias, CatalogAlias)
    assert catalog_alias.alias_path.is_symlink()
    assert catalog_alias.alias_path.exists()
    assert catalog_alias.alias_path.parent.name == CatalogInfix.ALIAS
    assert catalog_alias.target == Path("..") / CatalogInfix.ENTRY / (name + ".tgz")
    catalog_populated.assert_consistency()


def test_add_alias_unknown_name_raises(catalog_populated):
    with pytest.raises(AssertionError):
        catalog_populated.add_alias("nonexistent", "my-alias")


def test_add_alias_overwrite(catalog_populated):
    names = catalog_populated.list()
    name_a, name_b = names[0], names[1]
    alias = "shared-alias"

    catalog_populated.add_alias(name_a, alias)
    catalog_alias = catalog_populated.add_alias(name_b, alias)

    assert catalog_alias.catalog_entry.name == name_b
    assert catalog_alias.alias_path.is_symlink()
    assert (
        catalog_alias.alias_path.resolve()
        == (
            catalog_populated.repo_path / CatalogInfix.ENTRY / (name_b + ".tgz")
        ).resolve()
    )
    catalog_populated.assert_consistency()


def test_add_alias_multiple(catalog_populated):
    names = catalog_populated.list()
    aliases = [f"alias-{i}" for i in range(len(names))]

    for name, alias in zip(names, aliases):
        catalog_populated.add_alias(name, alias)

    catalog_aliases = catalog_populated.catalog_aliases
    assert len(catalog_aliases) == len(names)
    assert {ca.alias for ca in catalog_aliases} == set(aliases)
    catalog_populated.assert_consistency()


def test_add_alias_symlink_is_relative(catalog_populated):
    name = catalog_populated.list()[0]
    catalog_alias = catalog_populated.add_alias(name, "rel-alias")

    raw_target = Path(catalog_alias.alias_path.parent).joinpath(
        catalog_alias.alias_path.readlink()
    )
    assert not catalog_alias.alias_path.readlink().is_absolute()
    assert raw_target.resolve() == catalog_alias.alias_path.resolve()


def test_list_revisions_single(catalog_populated):
    name = catalog_populated.list()[0]
    catalog_alias = catalog_populated.add_alias(name, "rev-alias")

    revisions = catalog_alias.list_revisions()

    assert len(revisions) == 1
    entry, commit = revisions[0]
    assert isinstance(entry, CatalogEntry)
    assert entry.name == name
    assert commit.message.strip() == f"add alias: rev-alias -> {name}"


def test_list_revisions_overwrite(catalog_populated):
    names = catalog_populated.list()
    name_a, name_b = names[0], names[1]
    alias = "rev-alias"

    catalog_populated.add_alias(name_a, alias)
    catalog_alias = catalog_populated.add_alias(name_b, alias)

    revisions = catalog_alias.list_revisions()

    assert len(revisions) == 2
    # most recent first
    assert revisions[0][0].name == name_b
    assert revisions[1][0].name == name_a


def test_list_revisions_entries_require_exists_false(catalog_populated):
    names = catalog_populated.list()
    name_a, name_b = names[0], names[1]
    alias = "rev-alias"

    catalog_populated.add_alias(name_a, alias)
    catalog_alias = catalog_populated.add_alias(name_b, alias)
    catalog_populated.remove(name_a)

    revisions = catalog_alias.list_revisions()

    assert len(revisions) == 2
    entry_b, entry_a = revisions[0][0], revisions[1][0]
    assert entry_b.exists()
    assert not entry_a.exists()


def test_list_revisions_commit_objects(catalog_populated):
    name = catalog_populated.list()[0]
    catalog_alias = catalog_populated.add_alias(name, "rev-alias")

    revisions = catalog_alias.list_revisions()
    _, commit = revisions[0]

    assert hasattr(commit, "hexsha")
    assert hasattr(commit, "authored_datetime")
    assert hasattr(commit, "author")


def test_catalog_alias_from_name(catalog_populated):
    name = catalog_populated.list()[0]
    alias = "from-name-alias"
    catalog_populated.add_alias(name, alias)

    catalog_alias = CatalogAlias.from_name(alias, catalog_populated)

    assert isinstance(catalog_alias, CatalogAlias)
    assert catalog_alias.alias == alias
    assert catalog_alias.catalog_entry.name == name
    assert catalog_alias.alias_path.is_symlink()
    assert catalog_alias.catalog_entry.exists()


def test_catalog_alias_from_name_nonexistent_raises(catalog_populated):
    with pytest.raises(ValueError, match="no such alias"):
        CatalogAlias.from_name("does-not-exist", catalog_populated)


def test_catalog_alias_from_name_entry_consistency(catalog_populated):
    name = catalog_populated.list()[0]
    alias = "consistency-alias"
    catalog_populated.add_alias(name, alias)

    catalog_alias = CatalogAlias.from_name(alias, catalog_populated)

    catalog_alias.catalog_entry.assert_consistency()
    assert catalog_alias.catalog_entry.metadata_path.exists()
    assert catalog_alias.catalog_entry.catalog_path.exists()


def test_catalog_alias_from_name_matches_catalog_aliases(catalog_populated):
    name = catalog_populated.list()[0]
    alias = "match-alias"
    catalog_populated.add_alias(name, alias)

    from_name = CatalogAlias.from_name(alias, catalog_populated)
    from_catalog = next(
        ca for ca in catalog_populated.catalog_aliases if ca.alias == alias
    )

    assert from_name.alias == from_catalog.alias
    assert from_name.catalog_entry.name == from_catalog.catalog_entry.name


def test_catalog_entry_relocatable(repo_cloned_bare, tmpdir):
    cloned = Catalog.clone_from(
        repo_cloned_bare.working_dir, Path(tmpdir).joinpath("cloned")
    )
    catalog_entries = cloned.catalog_entries
    exprs = tuple(catalog_entry.expr for catalog_entry in catalog_entries)
    assert exprs



from __future__ import annotations

import functools
import inspect
import math
import operator
from collections import defaultdict
from datetime import date
from typing import TYPE_CHECKING, Any, NamedTuple
from urllib.parse import parse_qs, urlsplit
from uuid import uuid4

from cityhash import CityHash32


if TYPE_CHECKING:
    from collections.abc import Callable

try:
    import regex as re
except ImportError:
    import re


class _UDF(NamedTuple):
    """An internal record holding info about a registered UDF."""

    name: str
    impl: Any
    nargs: int
    skip_if_exists: bool = False
    deterministic: bool = True


_SQLITE_UDF_REGISTRY = {}
_SQLITE_UDAF_REGISTRY = {}
UNSET = object()


def ignore_nulls(f):
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        if any(arg is None for arg in args):
            return None
        return f(*args, **kwargs)

    return wrapper


def _number_of_arguments(callable):
    signature = inspect.signature(callable)
    parameters = signature.parameters.values()
    kinds = [param.kind for param in parameters]
    valid_kinds = (
        inspect.Parameter.POSITIONAL_OR_KEYWORD,
        inspect.Parameter.POSITIONAL_ONLY,
    )
    if any(kind not in valid_kinds for kind in kinds) or any(
        param.default is not inspect.Parameter.empty for param in parameters
    ):
        raise TypeError(
            "Only positional arguments without defaults are supported in Ibis "
            "SQLite function registration"
        )
    return len(parameters)


def udf(func=None, *, skip_if_exists=False, deterministic=True):
    """Create a SQLite scalar UDF from `func`.

    Parameters
    ----------
    func
        A callable object
    skip_if_exists
        If true, the UDF will only be registered if an existing function
        with that name doesn't already exist.
    deterministic
        Whether the UDF is deterministic, defaults to True.

    Returns
    -------
    callable
        A callable object that returns `None` if any of its inputs are
        `None`.

    """
    if func is None:
        return lambda func: udf(
            func, skip_if_exists=skip_if_exists, deterministic=deterministic
        )

    name = func.__name__
    nargs = _number_of_arguments(func)
    wrapper = ignore_nulls(func)

    _SQLITE_UDF_REGISTRY[name] = _UDF(
        name, wrapper, nargs, skip_if_exists, deterministic
    )
    return wrapper


def udaf(cls):
    """Register a UDAF class with any SQLite connection."""
    name = cls.__name__
    nargs = _number_of_arguments(cls.step) - 1
    _SQLITE_UDAF_REGISTRY[name] = _UDF(name, cls, nargs)
    return cls


# Optional builtin functions
#
# These functions may exist as builtins depending on the SQLite versions.
# They're only registered if they don't exist already in the connection.


@udf(skip_if_exists=True)
def unhex(string):
    return bytes.fromhex(string)


@udf(skip_if_exists=True)
def exp(arg):
    return math.exp(arg)


@udf(skip_if_exists=True)
def ln(arg):
    if arg < 0:
        return None
    return math.log(arg)


@udf(skip_if_exists=True)
def log(base, arg):
    """Return the logarithm of `arg` in the given `base`.

    The argument order matches the builtin sqlite function.
    """
    if arg < 0:
        return None
    return math.log(arg, base)


@udf(skip_if_exists=True)
def log2(arg):
    if arg < 0:
        return None
    return math.log2(arg)


@udf(skip_if_exists=True)
def log10(arg):
    if arg < 0:
        return None
    return math.log10(arg)


@udf(skip_if_exists=True)
def floor(arg):
    return math.floor(arg)


@udf(skip_if_exists=True)
def ceil(arg):
    return math.ceil(arg)


@udf(skip_if_exists=True)
def sign(arg):
    if not arg:
        return 0
    return math.copysign(1, arg)


@udf(skip_if_exists=True)
def mod(left, right):
    return None if right == 0 else (left % right)


@udf(skip_if_exists=True)
def power(arg, power):
    # mirroring sqlite - return NULL if negative or non-integral
    if arg < 0.0 and not power.is_integer():
        return None
    return arg**power


@udf(skip_if_exists=True)
def sqrt(arg):
    return None if arg < 0.0 else math.sqrt(arg)


@udf(skip_if_exists=True)
def sin(arg):
    return math.sin(arg)


@udf(skip_if_exists=True)
def cos(arg):
    return math.cos(arg)


@udf(skip_if_exists=True)
def tan(arg):
    return math.tan(arg)


@udf(skip_if_exists=True)
def asin(arg):
    return math.asin(arg)


@udf(skip_if_exists=True)
def acos(arg):
    return math.acos(arg)


@udf(skip_if_exists=True)
def atan(arg):
    return math.atan(arg)


@udf(skip_if_exists=True)
def atan2(y, x):
    return math.atan2(y, x)


@udf(skip_if_exists=True)
def degrees(x):
    return math.degrees(x)


@udf(skip_if_exists=True)
def radians(x):
    return math.radians(x)


@udf(skip_if_exists=True)
def pi():
    return math.pi


@udf(skip_if_exists=True, deterministic=False)
def uuid():
    return str(uuid4())


@udf(skip_if_exists=True, deterministic=True)
def city_hash_32(x):
    return CityHash32(str(x))


# Additional UDFS


@udf
def _ibis_reverse(string):
    return string[::-1]


@udf
def _ibis_string_ascii(string):
    return ord(string[0])


@udf
def _ibis_rpad(string, width, pad):
    return string.ljust(width, pad)[:width]


@udf
def _ibis_lpad(string, width, pad):
    return string.rjust(width, pad)[:width]


@udf
def _ibis_repeat(string, n):
    return string * n


@udf
def _ibis_translate(string, from_string, to_string):
    table = str.maketrans(from_string, to_string)
    return string.translate(table)


@udf
def _ibis_regex_search(string, regex):
    """Return whether `regex` exists in `string`."""
    return re.search(regex, string) is not None


@udf
def _ibis_regex_replace(string, pattern, replacement):
    """Replace occurrences of `pattern` in `string` with `replacement`."""
    return re.sub(pattern, replacement, string)


@udf
def _ibis_regex_extract(string, pattern, index):
    """Extract match of regular expression `pattern` from `string` at `index`."""
    result = re.search(pattern, string)
    if result is not None and 0 <= index <= (result.lastindex or -1):
        return result.group(index)
    return None


@udf
def _ibis_xor(x, y):
    return x ^ y


@udf
def _ibis_inv(x):
    return ~x


@udf
def _ibis_extract_fragment(url):
    return _extract_url_field(url, "fragment")


@udf
def _ibis_extract_protocol(url):
    return _extract_url_field(url, "scheme")


@udf
def _ibis_extract_authority(url):
    return _extract_url_field(url, "netloc")


@udf
def _ibis_extract_path(url):
    return _extract_url_field(url, "path")


@udf
def _ibis_extract_host(url):
    return _extract_url_field(url, "hostname")


def _extract_url_field(data, field_name):
    return getattr(urlsplit(data), field_name, "")


@udf
def _ibis_extract_full_query(url):
    return urlsplit(url).query


@udf
def _ibis_extract_query(url, param_name):
    query = urlsplit(url).query
    value = parse_qs(query)[param_name]
    return value if len(value) > 1 else value[0]


@udf
def _ibis_extract_user_info(url):
    url_parts = urlsplit(url)
    username = url_parts.username or ""
    password = url_parts.password or ""

    return f"{username}:{password}"


@udf
def _ibis_date_delta(left, right):
    delta = date.fromisoformat(left) - date.fromisoformat(right)
    return delta.days


class _ibis_var:
    def __init__(self, offset):
        self.mean = 0.0
        self.sum_of_squares_of_differences = 0.0
        self.count = 0
        self.offset = offset

    def step(self, value):
        if value is not None:
            self.count += 1
            delta = value - self.mean
            self.mean += delta / self.count
            self.sum_of_squares_of_differences += delta * (value - self.mean)

    def finalize(self):
        count = self.count
        if count:
            return self.sum_of_squares_of_differences / (count - self.offset)
        return None


@udaf
class _ibis_mode:
    def __init__(self):
        self.counter = defaultdict(int)

    def step(self, value):
        if value is not None:
            self.counter[value] += 1

    def finalize(self):
        if self.counter:
            return max(self.counter, key=self.counter.get)
        return None


@udaf
class _ibis_var_pop(_ibis_var):
    def __init__(self):
        super().__init__(0)


@udaf
class _ibis_var_sample(_ibis_var):
    def __init__(self):
        super().__init__(1)


class _ibis_bit_agg:
    def __init__(self, op):
        self.value: int | None = None
        self.count: int = 0
        self.op: Callable[[int, int], int] = op

    def step(self, value):
        if value is not None:
            if not self.count:
                self.value = value
            else:
                self.value = self.op(self.value, value)
            self.count += 1

    def finalize(self) -> int | None:
        return self.value


@udaf
class _ibis_bit_or(_ibis_bit_agg):
    def __init__(self):
        super().__init__(operator.or_)


@udaf
class _ibis_bit_and(_ibis_bit_agg):
    def __init__(self):
        super().__init__(operator.and_)


@udaf
class _ibis_bit_xor(_ibis_bit_agg):
    def __init__(self):
        super().__init__(operator.xor)


class _ibis_first_last:
    def __init__(self):
        self.value = None

    def finalize(self):
        return self.value


@udaf
class _ibis_first(_ibis_first_last):
    def step(self, value):
        if self.value is None:
            self.value = value


@udaf
class _ibis_first_include_null(_ibis_first_last):
    def __init__(self):
        self.value = UNSET

    def step(self, value):
        if self.value is UNSET:
            self.value = value


@udaf
class _ibis_last(_ibis_first_last):
    def step(self, value):
        if value is not None:
            self.value = value


@udaf
class _ibis_last_include_null(_ibis_first_last):
    def step(self, value):
        self.value = value


def register_all(con):
    """Register all udf and udaf with the connection.

    All udf and udaf are defined in this file with the `udf` and `udaf`
    decorators.
    """
    existing = {
        name for (name,) in con.execute("SELECT name FROM pragma_function_list()")
    }

    for udf in _SQLITE_UDF_REGISTRY.values():
        if udf.skip_if_exists and udf.name in existing:
            continue
        con.create_function(
            udf.name, udf.nargs, udf.impl, deterministic=udf.deterministic
        )

    for udf in _SQLITE_UDAF_REGISTRY.values():
        con.create_aggregate(udf.name, udf.nargs, udf.impl)



import re

import pytest

import xorq.api as xo
from xorq.caching import (
    ParquetCache,
    ParquetSnapshotCache,
    SourceCache,
    SourceSnapshotCache,
)


def test_into_backend(batting):
    xo.options.interactive = False

    ddb_con = xo.duckdb.connect()

    t = batting.filter(batting.yearID == 2015).into_backend(ddb_con, "ls_batting")

    expr = (
        t.join(t, "playerID")
        .limit(15)
        .select(player_id="playerID", year_id="yearID_right")
    )

    assert "RemoteTable[r1, name=ls_batting]" in repr(expr)


@pytest.mark.parametrize(
    "cache, strategy, parquet",
    [
        pytest.param(
            ParquetCache.from_kwargs(), "modification_time", True, id="parquet_cache"
        ),
        pytest.param(
            ParquetSnapshotCache.from_kwargs(),
            "snapshot",
            True,
            id="parquet_snapshot_cache",
        ),
        pytest.param(
            SourceCache.from_kwargs(), "modification_time", False, id="source_cache"
        ),
        pytest.param(
            SourceSnapshotCache.from_kwargs(),
            "snapshot",
            False,
            id="source_snapshot_cache",
        ),
    ],
)
def test_cache(batting, cache, strategy, parquet):
    xo.options.interactive = False

    expr = (
        batting.join(batting, "playerID")
        .limit(15)
        .select(player_id="playerID", year_id="yearID_right")
        .cache(cache)
    )

    pattern = rf"CachedNode\[r\d+, strategy={strategy}, parquet={parquet}, source"

    assert re.search(pattern, repr(expr))



from __future__ import annotations

import pytest
from pytest import param

from xorq.common.utils.ibis_utils import from_ibis
from xorq.tests.util import assert_frame_equal, assert_series_equal


ibis = pytest.importorskip("ibis")


@pytest.mark.parametrize(
    "window_fn",
    [
        param(
            lambda t, win: t.float_col.lag().over(win),
            id="lag",
        ),
        param(
            lambda t, win: t.float_col.lag(2).over(win),
            id="lag_offset",
        ),
        param(
            lambda t, win: t.float_col.lead().over(win),
            id="lead",
        ),
        param(
            lambda t, win: t.float_col.lead(3).over(win),
            id="lead_offset",
        ),
        param(
            lambda t, win: t.id.rank().over(win),
            id="rank",
        ),
        param(
            lambda t, win: t.id.dense_rank().over(win),
            id="dense_rank",
        ),
        param(
            lambda t, win: t.id.percent_rank().over(win),
            id="percent_rank",
        ),
        param(
            lambda t, win: t.id.cume_dist().over(win),
            id="cume_dist",
        ),
        param(
            lambda t, win: t.float_col.first().over(win),
            id="first",
        ),
        param(
            lambda t, win: t.float_col.last().over(win),
            id="last",
        ),
        param(
            lambda t, win: t.double_col.nth(3).over(win),
            id="nth",
        ),
        param(
            lambda _, win: ibis.row_number().over(win),
            id="row_number",
        ),
        param(
            lambda t, win: t.double_col.cumsum().over(win),
            id="cumsum",
        ),
        param(
            lambda t, win: t.double_col.cummean().over(win),
            id="cummean",
        ),
        param(
            lambda t, win: t.float_col.cummin().over(win),
            id="cummin",
        ),
        param(
            lambda t, win: t.float_col.cummax().over(win),
            id="cummax",
        ),
        param(
            lambda t, win: (t.double_col == 0).any().over(win),
            id="cumany",
        ),
        param(
            lambda t, win: (t.double_col == 0).notany().over(win),
            id="cumnotany",
        ),
        param(
            lambda t, win: (t.double_col == 0).all().over(win),
            id="cumall",
        ),
        param(
            lambda t, win: (t.double_col == 0).notall().over(win),
            id="cumnotall",
        ),
        param(
            lambda t, win: t.double_col.sum().over(win),
            id="sum",
        ),
        param(
            lambda t, win: t.double_col.mean().over(win),
            id="mean",
        ),
        param(
            lambda t, win: t.float_col.min().over(win),
            id="min",
        ),
        param(
            lambda t, win: t.float_col.max().over(win),
            id="max",
        ),
        param(
            lambda t, win: t.double_col.count().over(win),
            id="count",
        ),
    ],
)
def test_window_functions_grouped_ordered(ibis_alltypes, window_fn):
    win = ibis.window(
        group_by=[ibis_alltypes.string_col],
        order_by=[ibis_alltypes.id],
    )
    expr = ibis_alltypes.mutate(val=window_fn(ibis_alltypes, win))

    xorq_expr = from_ibis(expr)
    actual = xorq_expr.execute()
    expected = expr.execute()

    assert_frame_equal(
        expected.sort_values("id"), actual.sort_values("id"), check_dtype=False
    )


@pytest.mark.parametrize(
    "window_fn",
    [
        param(
            lambda t, win: ibis.row_number().over(win),
            id="row_number",
        ),
        param(
            lambda t, win: t.double_col.sum().over(win),
            id="sum",
        ),
        param(
            lambda t, win: t.double_col.mean().over(win),
            id="mean",
        ),
        param(
            lambda t, win: t.float_col.lag().over(win),
            id="lag",
        ),
        param(
            lambda t, win: t.float_col.lead().over(win),
            id="lead",
        ),
    ],
)
@pytest.mark.parametrize(
    "ordered",
    [
        param(True, id="ordered"),
        param(False, id="unordered"),
    ],
)
def test_window_functions_ungrouped(ibis_alltypes, window_fn, ordered):
    order_by = [ibis_alltypes.id] if ordered else None
    win = ibis.window(order_by=order_by)
    expr = ibis_alltypes.mutate(val=window_fn(ibis_alltypes, win))

    xorq_expr = from_ibis(expr)

    actual = xorq_expr.execute()
    expected = expr.execute()
    assert_frame_equal(
        expected.sort_values("id"), actual.sort_values("id"), check_dtype=False
    )


@pytest.mark.parametrize(
    ("preceding", "following"),
    [
        param(2, 0, id="preceding_2_following_0"),
        param((2, 0), None, id="preceding_tuple"),
        param(0, 2, id="preceding_0_following_2"),
        param(None, (0, 2), id="following_tuple"),
        param(1, 1, id="preceding_1_following_1"),
    ],
)
def test_window_frame_rows_bounded(ibis_alltypes, preceding, following):
    win = ibis.window(
        preceding=preceding,
        following=following,
        group_by=[ibis_alltypes.string_col],
        order_by=[ibis_alltypes.id],
    )
    expr = ibis_alltypes.mutate(val=ibis_alltypes.double_col.sum().over(win))

    xorq_expr = from_ibis(expr)

    expected = expr.execute()
    actual = xorq_expr.execute()
    assert_frame_equal(
        expected.sort_values("id"), actual.sort_values("id"), check_dtype=False
    )


def test_window_frame_trailing(ibis_alltypes):
    win = ibis.trailing_window(
        preceding=2,
        group_by=[ibis_alltypes.string_col],
        order_by=[ibis_alltypes.id],
    )
    expr = ibis_alltypes.mutate(val=ibis_alltypes.double_col.sum().over(win))

    xorq_expr = from_ibis(expr)

    expected = expr.execute()
    actual = xorq_expr.execute()
    assert_frame_equal(
        expected.sort_values("id"), actual.sort_values("id"), check_dtype=False
    )


def test_window_frame_unbounded_following(ibis_alltypes):
    win = ibis.window(
        rows=(0, None),
        order_by=[ibis_alltypes.id],
    )
    expr = ibis_alltypes.mutate(val=ibis_alltypes.double_col.sum().over(win))

    xorq_expr = from_ibis(expr)

    expected = expr.execute()
    actual = xorq_expr.execute()
    assert_frame_equal(
        expected.sort_values("id"), actual.sort_values("id"), check_dtype=False
    )


def test_window_frame_unbounded_preceding(ibis_alltypes):
    win = ibis.window(
        following=0,
        group_by=[ibis_alltypes.string_col],
        order_by=[ibis_alltypes.id],
    )
    expr = ibis_alltypes.mutate(val=ibis_alltypes.double_col.sum().over(win))

    xorq_expr = from_ibis(expr)

    expected = expr.execute()
    actual = xorq_expr.execute()
    assert_frame_equal(
        expected.sort_values("id"), actual.sort_values("id"), check_dtype=False
    )


def test_window_range_specification(ibis_alltypes):
    win = ibis.range_window(
        preceding=10,
        following=0,
        order_by="id",
        group_by="string_col",
    )
    expr = ibis_alltypes.mutate(val=ibis_alltypes.double_col.sum().over(win))

    xorq_expr = from_ibis(expr)

    expected = expr.execute()
    actual = xorq_expr.execute()
    assert_frame_equal(
        expected.sort_values("id"), actual.sort_values("id"), check_dtype=False
    )


@pytest.mark.parametrize(
    "group_by_cols",
    [
        param([ibis._.string_col], id="single_column"),
        param([ibis._.string_col, ibis._.bool_col], id="multiple_columns"),
    ],
)
def test_window_partition_by(ibis_alltypes, group_by_cols):
    win = ibis.window(
        group_by=group_by_cols,
        order_by=[ibis_alltypes.id],
    )
    expr = ibis_alltypes.mutate(val=ibis.row_number().over(win))

    xorq_expr = from_ibis(expr)

    expected = expr.execute()
    actual = xorq_expr.execute()
    assert_frame_equal(
        expected.sort_values("id"), actual.sort_values("id"), check_dtype=False
    )


@pytest.mark.parametrize(
    "order_by_spec",
    [
        param([ibis._.id], id="single_asc"),
        param([ibis.desc(ibis._.id)], id="single_desc"),
        param([ibis._.string_col, ibis._.id], id="multiple_asc"),
        param(
            [ibis.asc(ibis._.string_col), ibis.desc(ibis._.id)],
            id="multiple_mixed",
        ),
    ],
)
def test_window_order_by(ibis_alltypes, order_by_spec):
    win = ibis.window(
        group_by=[ibis_alltypes.string_col],
        order_by=order_by_spec,
    )
    expr = ibis_alltypes.mutate(val=ibis.row_number().over(win))

    xorq_expr = from_ibis(expr)

    expected = expr.execute()
    actual = xorq_expr.execute()
    assert_frame_equal(
        expected.sort_values("id"), actual.sort_values("id"), check_dtype=False
    )


def test_window_ntile(ibis_alltypes):
    win = ibis.window(
        group_by=[ibis_alltypes.string_col],
        order_by=[ibis_alltypes.id],
    )

    expr = ibis_alltypes.mutate(val=ibis.ntile(7).over(win))

    xorq_expr = from_ibis(expr)
    assert xorq_expr is not None
    result = xorq_expr.execute()
    assert not result.empty


def test_window_multiple_functions(ibis_alltypes):
    win = ibis.window(
        group_by=[ibis_alltypes.string_col],
        order_by=[ibis_alltypes.id],
    )
    expr = ibis_alltypes.mutate(
        row_num=ibis.row_number().over(win),
        running_sum=ibis_alltypes.double_col.sum().over(win),
        running_avg=ibis_alltypes.double_col.mean().over(win),
        prev_val=ibis_alltypes.float_col.lag().over(win),
    )

    xorq_expr = from_ibis(expr)

    expected = expr.execute()
    actual = xorq_expr.execute()
    assert_frame_equal(
        expected.sort_values("id"), actual.sort_values("id"), check_dtype=False
    )


def test_window_with_filter(ibis_alltypes):
    win = ibis.window(order_by=[ibis_alltypes.id])
    expr = (
        ibis_alltypes.mutate(next_int=ibis_alltypes.int_col.lead().over(win))
        .filter(lambda t: t.int_col == 1)
        .select("int_col", "next_int")
    )

    xorq_expr = from_ibis(expr)

    expected = expr.execute()
    actual = xorq_expr.execute()
    assert_frame_equal(expected, actual, check_dtype=False)


def test_window_coalesce(ibis_alltypes):
    expr = (
        ibis_alltypes.group_by("month")
        .order_by("id")
        .mutate(lagged_value=ibis.coalesce(ibis_alltypes.bigint_col.lag(), 0))
        .select(["id", "lagged_value"])
    )

    xorq_expr = from_ibis(expr)

    expected = expr.execute()
    actual = xorq_expr.execute()
    assert_frame_equal(
        expected.sort_values("id"), actual.sort_values("id"), check_dtype=False
    )


def test_window_percent_rank_no_order(ibis_alltypes):
    expr = ibis_alltypes.mutate(val=ibis_alltypes.id.percent_rank())

    xorq_expr = from_ibis(expr)

    expected = expr.execute()
    actual = xorq_expr.execute()
    assert_frame_equal(
        expected.sort_values("id"), actual.sort_values("id"), check_dtype=False
    )


def test_window_rank_over_call(ibis_alltypes):
    expr = ibis_alltypes.int_col.percent_rank().over(
        ibis.window(group_by=ibis_alltypes.int_col.notnull())
    )

    xorq_expr = from_ibis(expr)

    expected = expr.execute()
    actual = xorq_expr.execute()
    assert_series_equal(expected.sort_values(), actual.sort_values(), check_dtype=False)


def test_window_order_by_with_nulls(ibis_alltypes):
    win = ibis.window(
        order_by=[
            ibis.asc(ibis_alltypes.bool_col.isnull()),
            ibis.asc(ibis_alltypes.bool_col),
        ],
    )
    expr = ibis_alltypes.select(
        rank=ibis_alltypes.bool_col.rank().over(win),
        bool_col=ibis_alltypes.bool_col,
    )

    xorq_expr = from_ibis(expr)

    expected = expr.execute()
    actual = xorq_expr.execute()
    assert_frame_equal(expected, actual, check_dtype=False)


def test_window_cumsum_shorthand(ibis_alltypes):
    expr = ibis_alltypes.mutate(csum=ibis_alltypes.double_col.cumsum(order_by="id"))

    xorq_expr = from_ibis(expr)

    expected = expr.execute()
    actual = xorq_expr.execute()
    assert_frame_equal(
        expected.sort_values("id"), actual.sort_values("id"), check_dtype=False
    )


@pytest.mark.parametrize(
    ("window_fn", "aggr_fn"),
    [
        param(
            lambda t: ibis.window(group_by=[t.string_col], order_by=[t.id]),
            lambda t: t.double_col.sum(),
            id="sum",
        ),
        param(
            lambda t: ibis.window(group_by=[t.string_col], order_by=[t.id]),
            lambda t: t.double_col.mean(),
            id="mean",
        ),
        param(
            lambda t: ibis.window(group_by=[t.string_col], order_by=[t.id]),
            lambda t: t.double_col.min(),
            id="min",
        ),
        param(
            lambda t: ibis.window(group_by=[t.string_col], order_by=[t.id]),
            lambda t: t.double_col.max(),
            id="max",
        ),
        param(
            lambda t: ibis.window(group_by=[t.string_col], order_by=[t.id]),
            lambda t: t.double_col.count(),
            id="count",
        ),
        param(
            lambda t: ibis.window(group_by=[t.string_col], order_by=[t.id]),
            lambda t: t.double_col.std(),
            id="std",
        ),
        param(
            lambda t: ibis.window(group_by=[t.string_col], order_by=[t.id]),
            lambda t: t.double_col.var(),
            id="var",
        ),
    ],
)
def test_window_aggregate_functions(ibis_alltypes, window_fn, aggr_fn):
    win = window_fn(ibis_alltypes)
    expr = ibis_alltypes.mutate(val=aggr_fn(ibis_alltypes).over(win))

    xorq_expr = from_ibis(expr)

    expected = expr.execute()
    actual = xorq_expr.execute()
    assert_frame_equal(
        expected.sort_values("id"), actual.sort_values("id"), check_dtype=False
    )
